import ntpath
import os
import posixpath
from enum import Enum
from itertools import chain
from six import string_types

from .. import safe_str
from ..objutils import objectify

Root = Enum('Root', ['srcdir', 'builddir', 'absolute'])
InstallRoot = Enum('InstallRoot', ['prefix', 'exec_prefix', 'bindir', 'libdir',
                                   'includedir'])
DestDir = Enum('DestDir', ['destdir'])


class BasePath(safe_str.safe_string):
    __slots__ = ['destdir', 'root', 'suffix']

    curdir = posixpath.curdir
    pardir = posixpath.pardir
    sep = posixpath.sep

    __repr_variables = dict(
        [(i, '$({})'.format(i.name)) for i in chain(Root, InstallRoot)] +
        [(DestDir.destdir, '$(DESTDIR)')]
    )

    def __init__(self, path, root=Root.builddir, destdir=False):
        if destdir and root not in InstallRoot:
            raise ValueError('destdir only applies to install paths')
        self.destdir = destdir

        drive, path = self.__normalize(path, expand_user=True)
        self.suffix = drive + path

        if posixpath.isabs(path):
            self.root = Root.absolute
        elif root == Root.absolute:
            raise ValueError("'{}' is not absolute".format(path))
        else:
            self.root = root

    @classmethod
    def abspath(cls, path):
        drive, path = cls.__normalize(path, expand_user=True)
        cwddrive, cwdpath = cls.__normalize(os.getcwd())

        if not drive:
            drive = cwddrive
        path = posixpath.normpath(posixpath.join(cwdpath, path))
        return cls(drive + path, Root.absolute)

    @classmethod
    def ensure(cls, path, root=Root.builddir, destdir=False, base=None,
               strict=False):
        result = objectify(path, base or cls, cls, root=root, destdir=destdir)
        if strict and result.root != root:
            raise ValueError('expected root of {!r}, but got {!r}'
                             .format(root.name, result.root.name))
        return result

    @staticmethod
    def __normalize(path, expand_user=False):
        if expand_user:
            path = os.path.expanduser(path)
        drive, path = ntpath.splitdrive(path)
        if drive and not ntpath.isabs(path):
            raise ValueError('relative paths with drives not supported')

        drive = drive.replace('\\', '/')
        path = posixpath.normpath(path.replace('\\', '/'))
        if path == '.':
            path = ''
        return drive, path

    def __localize(self, thing):
        if isinstance(thing, string_types):
            return self._localize_path(thing)
        return thing

    def cross(self, env):
        cls = env.target_platform.Path
        return cls(self.suffix, self.root)

    def parent(self):
        if not self.suffix:
            raise ValueError('already at root')
        return type(self)(posixpath.dirname(self.suffix), self.root,
                          self.destdir)

    def append(self, path):
        drive, path = self.__normalize(path)
        if not posixpath.isabs(path):
            path = posixpath.normpath(posixpath.join(self.suffix, path))
        return type(self)(drive + path, self.root, self.destdir)

    def ext(self):
        return posixpath.splitext(self.suffix)[1]

    def addext(self, ext):
        return type(self)(self.suffix + ext, self.root, self.destdir)

    def stripext(self, replace=None):
        name = posixpath.splitext(self.suffix)[0]
        if replace:
            name += replace
        return type(self)(name, self.root, self.destdir)

    def splitleaf(self):
        return self.parent(), self.basename()

    def split(self):
        # This is guaranteed to work since `suffix` is normalized.
        return self.suffix.split(posixpath.sep)

    def basename(self):
        return posixpath.basename(self.suffix)

    def relpath(self, start, prefix=''):
        if self.root == Root.absolute:
            return self.__localize(self.suffix)
        if self.root != start.root:
            raise ValueError('source mismatch')

        rel = posixpath.relpath(self.suffix or posixpath.curdir,
                                start.suffix or posixpath.curdir)
        if prefix and rel == self.curdir:
            return prefix
        return self.__localize(posixpath.join(prefix, rel))

    def reroot(self, root=Root.builddir):
        return type(self)(self.suffix, root, self.destdir)

    def to_json(self):
        return (self.suffix, self.root.name, self.destdir)

    @classmethod
    def from_json(cls, data):
        try:
            base = Root[data[1]]
        except KeyError:
            base = InstallRoot[data[1]]
        return cls(data[0], base, data[2])

    def realize(self, variables, executable=False):
        root = variables[self.root] if self.root != Root.absolute else None
        if executable and root is None and posixpath.sep not in self.suffix:
            root = posixpath.curdir

        # Not all platforms (e.g. Windows) support $(DESTDIR), so only emit the
        # destdir variable if it's defined.
        if self.destdir and DestDir.destdir in variables:
            root = variables[DestDir.destdir] + root
        if root is None:
            return self.__localize(self.suffix or posixpath.curdir)
        if not self.suffix:
            return self.__localize(root)

        # Join the separator and the suffix first so that we don't end up with
        # unnecessarily-escaped backslashes on Windows. (It doesn't hurt
        # anything; it just looks weird.)
        return (self.__localize(root) +
                self.__localize(posixpath.sep + self.suffix))

    def string(self, variables=None):
        path = self
        result = ''

        while True:
            real = path.realize(variables)
            if isinstance(real, safe_str.jbos):
                path, suffix = real.bits
                result = suffix + result
            elif isinstance(real, BasePath):
                path = real
            else:
                result = real + result
                break

        return result

    def __str__(self):
        raise NotImplementedError()

    def __repr__(self):
        return '`{}`'.format(self.realize(self.__repr_variables))

    def __hash__(self):
        return hash(self.suffix)

    def __eq__(self, rhs):
        if type(self) is not type(rhs):
            return NotImplemented
        return (self.root == rhs.root and self.suffix == rhs.suffix and
                self.destdir == rhs.destdir)

    def __ne__(self, rhs):
        return not (self == rhs)



import errno
import os
import re
import subprocess
import sys

from .app_version import version
from .arguments import parser as argparse

# This captures a few kinds of messages emitted via `javac -verbose` (or
# scalac or whatever):
#   JDK 1-8:  [wrote RegularFileObject[path]]
#   JDK 9-10: [wrote DirectoryFileObject[bad:path]]
#   JDK 11+:  [wrote path]
#   Scala:    [wrote 'thing' to path]
_class_re = re.compile(
    r"\[wrote ((?:Regular|Directory)FileObject\[|'[^']*' to |)([^\]]*)"
)


def main():
    parser = argparse.ArgumentParser(
        prog='bfg9000-jvmoutput',
        description=('Read verbose output from a JVM compiler and write the ' +
                     'generated .class files to a file.')
    )
    parser.add_argument('--version', action='version',
                        version='%(prog)s ' + version)
    parser.add_argument('-o', type=argparse.FileType('w'), default=sys.stdout,
                        dest='output', help=('the output file to list the ' +
                                             'generated .class files'))
    parser.add_argument('command', nargs=argparse.REMAINDER, metavar='COMMAND',
                        help='the command to execute')
    args = parser.parse_args()

    if len(args.command) == 0:
        parser.error('command required')

    try:
        p = subprocess.Popen(args.command, universal_newlines=True,
                             stderr=subprocess.PIPE)
    except OSError as e:
        if e.errno == errno.ENOENT:
            parser.exit(66, 'command not found: {}\n'.format(args.command[0]))
        raise  # pragma: no cover

    for line in p.stderr:
        if line[0] != '[':
            sys.stderr.write(line)
            continue

        m = _class_re.match(line)
        if m:
            kind = m.group(1)
            class_file = m.group(2)
            # JDK 9-10 emits broken paths; fix them. For more info, see
            # <https://bugs.openjdk.java.net/browse/JDK-8194893>.
            if kind == 'DirectoryFileObject[':
                class_file = class_file.replace(':', os.path.sep)
            args.output.write(class_file + '\n')
    return p.wait()



import os
import re

from ... import path
from ... import shell
from .syntax import *
from ...iterutils import listify, uniques
from ...versioning import Version


def version(env=os.environ):
    try:
        make = shell.which(env.get('MAKE', ['make', 'gmake']), env)
        output = shell.execute(make + ['--version'], stdout=shell.Mode.pipe,
                               stderr=shell.Mode.devnull)
        m = re.match(r'GNU Make ([\d\.]+)', output)
        if m:
            return Version(m.group(1))
    except (IOError, OSError, shell.CalledProcessError):
        pass
    return None


priority = 2
filepath = path.Path('Makefile')

_rule_handlers = {}
_pre_rules = []
_post_rules = []

dir_sentinel = '.dir'


def rule_handler(*args):
    def decorator(fn):
        for i in args:
            _rule_handlers[i] = fn
        return fn
    return decorator


def pre_rule(fn):
    _pre_rules.append(fn)
    return fn


def post_rule(fn):
    _post_rules.append(fn)
    return fn


def write(env, build_inputs):
    buildfile = Makefile(build_inputs.bfgpath.string(env.base_dirs),
                         env.backend_version is not None)
    buildfile.variable(path_vars[path.Root.srcdir], env.srcdir, Section.path)

    for i in _pre_rules:
        i(build_inputs, buildfile, env)
    for e in build_inputs.edges():
        _rule_handlers[type(e)](e, build_inputs, buildfile, env)
    for i in _post_rules:
        i(build_inputs, buildfile, env)

    with open(filepath.string(env.base_dirs), 'w') as out:
        buildfile.write(out)


def flags_vars(name, value, buildfile):
    name = name.upper()
    gflags = buildfile.variable('GLOBAL_' + name, value, Section.flags, True)
    flags = buildfile.target_variable(name, gflags, True)
    return gflags, flags


def _get_path(thing):
    return thing if isinstance(thing, path.Path) else thing.path


def multitarget_rule(buildfile, targets, deps=None, order_only=None,
                     recipe=None, variables=None, phony=None):
    targets = listify(targets)
    if len(targets) > 1:
        first = targets[0]
        primary = _get_path(first).addext('.stamp')
        buildfile.rule(target=targets, deps=[primary])
        recipe = listify(recipe) + [Silent([ 'touch', qvar('@') ])]
    else:
        primary = targets[0]

    buildfile.rule(primary, deps, order_only, recipe, variables, phony)


def directory_deps(targets):
    builddir = path.Path('.')
    dirs = uniques(_get_path(i).parent() for i in targets)
    return [i.append(dir_sentinel) for i in dirs if i != builddir]


@post_rule
def directory_rule(build_inputs, buildfile, env):
    mkdir_p = env.tool('mkdir_p')
    pattern = Pattern(os.path.join('%', dir_sentinel))
    path = Function('patsubst', pattern, Pattern('%'), var('@'), quoted=True)

    buildfile.rule(
        target=pattern,
        recipe=[
            Silent(mkdir_p(path)),
            Silent(['touch', qvar('@')])
        ]
    )



import json
import uuid
from six import iteritems

from ... import path

__all__ = ['SlnBuilder', 'SlnElement', 'SlnVariable', 'Solution', 'UuidMap']


class SlnElement(object):
    def __init__(self, name, arg=None, value=None):
        if (arg is None) != (value is None):
            raise TypeError('if arg is passed, value must be too')
        self.name = name
        self.arg = arg
        self.value = value
        self.children = []

    def __call__(self, *args):
        return self.extend(args)

    def append(self, item):
        self.children.append(item)
        return self

    def extend(self, args):
        self.children.extend(args)
        return self

    def write(self, out, depth=0):
        out.write('\t' * depth)
        out.write(self.name)
        if self.arg:
            out.write('({}) = {}'.format(self.arg, self.value))
        out.write('\n')

        for i in self.children:
            i.write(out, depth + 1)

        out.write('\t' * depth + 'End' + self.name + '\n')


class SlnVariable(object):
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def write(self, out, depth=0):
        out.write('\t' * depth + '{} = {}\n'.format(self.name, self.value))


class SlnBuilder(object):
    def __init__(self):
        pass

    def __call__(self, *args, **kwargs):
        return SlnElement(*args, **kwargs)

    def __getattribute__(self, name):
        def closure(*args, **kwargs):
            return SlnElement(name, *args, **kwargs)
        return closure


class Solution(object):
    def __init__(self, uuids):
        self.uuid = uuids['']
        self._uuids = uuids
        self._projects = []
        self._project_map = {}

    def __setitem__(self, key, value):
        value.set_uuid(self._uuids)
        self._projects.append(value)
        self._project_map[key] = value

    def __getitem__(self, key):
        return self._project_map[key]

    def __iter__(self):
        return iter(self._projects)

    def __contains__(self, key):
        return key in self._project_map

    def dependencies(self, deps):
        # By definition, a dependency for an edge must already be defined by
        # the time the edge is created, so we can map *all* the dependencies to
        # their associated projects by looking at the projects we've already
        # created.
        dependencies = []
        for dep in deps:
            if not dep.creator:
                continue

            dep_output = dep.creator.output[0]
            if dep_output not in self:
                raise RuntimeError('unknown dependency for {!r}'.format(dep))
            dependencies.append(self[dep_output])
        return dependencies

    @property
    def uuid_str(self):
        return uuid_str(self.uuid)

    def write(self, out):
        S = SlnBuilder()
        Var = SlnVariable

        out.write('Microsoft Visual Studio Solution File, Format Version ' +
                  '12.00\n')
        out.write('# Visual Studio 14\n')

        Var('VisualStudioVersion', '14.0.22609.0').write(out)
        Var('MinimumVisualStudioVersion', '10.0.40219.1').write(out)

        configs = set()
        project_info = []

        for p in self._projects:
            path_vars = {path.Root.builddir: None}
            proj = S.Project(
                '"{}"'.format(self.uuid_str),
                '"{name}", "{path}", "{uuid}"'.format(
                    name=p.name, path=p.path.string(path_vars), uuid=p.uuid_str
                )
            )
            if p.dependencies:
                proj.append(
                    S.ProjectSection('ProjectDependencies', 'postProject')
                     .extend(Var(i.uuid_str, i.uuid_str)
                             for i in p.dependencies)
                )
            proj.write(out)

            configs.add(p.config_plat)
            project_info.append(Var('{uuid}.{cfg}.ActiveCfg'.format(
                uuid=p.uuid_str, cfg=p.config_plat
            ), p.real_config_plat))
            project_info.append(Var('{uuid}.{cfg}.Build.0'.format(
                uuid=p.uuid_str, cfg=p.config_plat
            ), p.real_config_plat))

        S.Global()(
            S.GlobalSection('SolutionConfigurationPlatforms', 'preSolution')
             .extend(Var(i, i) for i in configs),
            S.GlobalSection('ProjectConfigurationPlatforms', 'postSolution')
             .extend(project_info),
            S.GlobalSection('SolutionProperties', 'preSolution')(
                Var('HideSolutionNode', 'FALSE')
            )
        ).write(out)


def uuid_str(uuid):
    return '{{{}}}'.format(str(uuid).upper())


class UuidMap(object):
    version = 1

    def __init__(self, path):
        self._path = path
        self._seen = set()
        try:
            self._map = self._load(path)
        except IOError:
            self._map = {}

    def __getitem__(self, key):
        self._seen.add(key)
        if key in self._map:
            return self._map[key]
        else:
            u = uuid.uuid4()
            self._map[key] = u
            return u

    @classmethod
    def _load(cls, path):
        with open(path) as inp:
            state = json.load(inp)
        if state['version'] > cls.version:
            raise ValueError('saved version exceeds expected version')
        return { k: uuid.UUID(hex=v) for k, v in iteritems(state['map']) }

    def save(self, path=None):
        with open(path or self._path, 'w') as out:
            # Only save the UUIDs we saw this time. Skip ones we didn't see.
            seenmap = { k: v.hex for k, v in iteritems(self._map)
                        if k in self._seen }
            json.dump({
                'version': self.version,
                'map': seenmap,
            }, out)



from . import tool
from .common import not_buildroot, SimpleCommand

from ..iterutils import iterate


@tool('doppel')
class Doppel(SimpleCommand):
    def __init__(self, env):
        SimpleCommand.__init__(self, env, name='doppel', env_var='DOPPEL',
                               default='doppel')

    @property
    def data_args(self):
        return ['-m', '644']

    def _call(self, cmd, mode, src, dst, directory=None, format=None,
              dest_prefix=None):
        if mode == 'onto':
            return cmd + ['-p', src, dst]

        elif mode == 'into':
            result = cmd + ['-ipN']
            if not_buildroot(directory):
                result.extend(['-C', directory])
            result.extend(iterate(src))
            result.append(dst)
            return result

        elif mode == 'archive':
            result = cmd + ['-ipN', '-f', format]
            if not_buildroot(directory):
                result.extend(['-C', directory])
            if dest_prefix:
                result.extend(['-P', dest_prefix])
            result.extend(iterate(src))
            result.append(dst)
            return result

        raise ValueError("unknown mode '{}'".format(mode))



from collections import OrderedDict

from . import builtin
from ..iterutils import iterate
from ..backends.make import writer as make
from ..backends.ninja import writer as ninja
from ..path import Path, Root

_exts = OrderedDict(
    gzip='.tar.gz',
    bzip2='.tar.bz2',
    zip='.zip',
)


@builtin.function('builtins')
def extra_dist(builtins, files=None, dirs=None):
    for i in iterate(files):
        builtins['generic_file'](i)
    for i in iterate(dirs):
        builtins['directory'](i, include='*')


def _dist_command(format, build_inputs, buildfile, env):
    srcdir = Path('.', Root.srcdir)
    doppel = env.tool('doppel')

    project = build_inputs['project']
    dstname = project.name
    if project.version:
        dstname += '-' + str(project.version)

    return doppel(
        'archive', [i.path.relpath(srcdir) for i in build_inputs.sources()],
        Path(dstname + _exts[format]), directory=srcdir, format=format,
        dest_prefix=dstname
    )


@make.post_rule
def make_dist_rule(build_inputs, buildfile, env):
    for fmt in _exts:
        buildfile.rule(
            target='dist-{}'.format(fmt),
            recipe=[_dist_command(fmt, build_inputs, buildfile, env)],
            phony=True
        )

    buildfile.rule(
        target='dist',
        deps='dist-gzip',
        phony=True
    )


@ninja.post_rule
def ninja_dist_rule(build_inputs, buildfile, env):
    for fmt in _exts:
        ninja.command_build(
            buildfile, env,
            output='dist-{}'.format(fmt),
            command=_dist_command(fmt, build_inputs, buildfile, env),
            phony=True
        )

    buildfile.build(
        output='dist',
        rule='phony',
        inputs='dist-gzip'
    )



import os.path
import shutil

from . import *


class TestDepfile(IntegrationTest):
    def __init__(self, *args, **kwargs):
        IntegrationTest.__init__(self, 'depfile', stage_src=True,
                                 *args, **kwargs)

    @skip_pred(lambda x: x.backend == 'make' and
               env.host_platform.family == 'windows',
               'xfail on windows + make')
    def test_build(self):
        self.build(executable('program'))
        self.assertOutput([executable('program')], 'hello\n')

        self.wait()
        shutil.copy(os.path.join(self.srcdir, 'header_replaced.hpp'),
                    os.path.join(self.srcdir, 'header.hpp'))

        self.build(executable('program'))
        self.assertOutput([executable('program')], 'goodbye\n')



import os
import tempfile
import shutil

from . import *
pjoin = os.path.join


class TestInstall(IntegrationTest):
    def __init__(self, *args, **kwargs):
        IntegrationTest.__init__(self, 'install', install=True, *args,
                                 **kwargs)

    def test_default(self):
        self.build()
        self.assertOutput(
            [executable('program')],
            'hello from shared a!\nhello from shared b!\n' +
            'hello from static a!\nhello from static b!\n'
        )

    def _check_installed(self):
        self.build('install')

        extra = []
        if env.target_platform.has_import_library:
            extra = [pjoin(self.libdir, import_library('shared_a').path)]

        self.assertDirectory(self.installdir, [
            pjoin(self.includedir, 'shared_a.hpp'),
            pjoin(self.includedir, 'static_a.hpp'),
            pjoin(self.bindir, executable('program').path),
            pjoin(self.libdir, shared_library('shared_a').path),
            pjoin(self.libdir, shared_library('shared_b').path),
            pjoin(self.libdir, static_library('static_a').path),
        ] + extra)

    @skip_if_backend('msbuild')
    def test_install(self):
        self.build('install')
        self._check_installed()

        os.chdir(self.srcdir)
        cleandir(self.builddir)
        self.assertOutput(
            [pjoin(self.bindir, executable('program').path)],
            'hello from shared a!\nhello from shared b!\n' +
            'hello from static a!\nhello from static b!\n'
        )

    @skip_if_backend('msbuild')
    def test_install_existing_paths(self):
        makedirs(self.includedir, exist_ok=True)
        makedirs(self.bindir, exist_ok=True)
        makedirs(self.libdir, exist_ok=True)
        self.build('install')
        self._check_installed()

        os.chdir(self.srcdir)
        cleandir(self.builddir)
        self.assertOutput(
            [pjoin(self.bindir, executable('program').path)],
            'hello from shared a!\nhello from shared b!\n' +
            'hello from static a!\nhello from static b!\n'
        )

    @skip_if_backend('msbuild')
    def test_uninstall(self):
        self.build('install')
        self._check_installed()

        self.build('uninstall')
        self.assertDirectory(self.installdir, [])


# No DESTDIR on Windows.
@skip_if(env.host_platform.family == 'windows', hide=True)
class TestDestDir(IntegrationTest):
    def __init__(self, *args, **kwargs):
        IntegrationTest.__init__(self, 'install', install=True,
                                 configure=False, *args, **kwargs)

    def setUp(self):
        self.destdir = tempfile.mkdtemp(prefix='destdir-')

    def tearDown(self):
        cleandir(self.destdir, recreate=False)

    def _check_installed(self):
        self.assertDirectory(self.destdir + self.installdir, [
            self.destdir + pjoin(self.includedir, 'shared_a.hpp'),
            self.destdir + pjoin(self.includedir, 'static_a.hpp'),
            self.destdir + pjoin(self.bindir, executable('program').path),
            self.destdir + pjoin(self.libdir, shared_library('shared_a').path),
            self.destdir + pjoin(self.libdir, shared_library('shared_b').path),
            self.destdir + pjoin(self.libdir, static_library('static_a').path),
        ])

    def _check_run(self):
        path = os.path.normpath(self.destdir + self.installdir)
        for i in os.listdir(path):
            shutil.move(pjoin(path, i), self.installdir)
        self.assertOutput(
            [pjoin(self.bindir, executable('program').path)],
            'hello from shared a!\nhello from shared b!\n' +
            'hello from static a!\nhello from static b!\n'
        )

    @skip_if_backend('msbuild')
    def test_install_destdir(self):
        self.configure(env={'DESTDIR': self.destdir})
        self.build('install')
        self._check_installed()
        self._check_run()

    @only_if_backend('make', hide=True)
    def test_install_override_destdir(self):
        self.configure()
        self.build('install', extra_args=['DESTDIR={}'.format(self.destdir)])
        self._check_installed()
        self._check_run()

    @skip_if_backend('msbuild')
    def test_uninstall_destdir(self):
        self.configure(env={'DESTDIR': self.destdir})
        self.build('install')
        self._check_installed()

        self.build('uninstall')
        self.assertDirectory(self.destdir + self.installdir, [])



from .core import _get_platform_info, _platform_info, Platform


class TargetPlatform(Platform):
    _package_map = {}

    def transform_package(self, name):
        return self._package_map.get(name, name)


def platform_info(*args, **kwargs):
    return _platform_info('target', *args, **kwargs)


def from_json(value):
    return _get_platform_info('target', value['genus'], value['species'],
                              value['arch'])



import mock
import ntpath
import os.path
import posixpath
import sys
from six import iteritems

from .. import *

from bfg9000.path import Path
from bfg9000.platforms.posix import PosixPath
from bfg9000.platforms.windows import WindowsPath

if sys.version_info >= (3,):
    mock_open_name = 'builtins.open'
else:
    mock_open_name = '__builtin__.open'


# Fix the mock package's mock_open function to work with iter(); note: this is
# already fixed in Python 3.7.1's unittest.mock.
def mock_open(*args, **kwargs):
    mo = mock.mock_open(*args, **kwargs)
    handle = mo.return_value
    handle.__iter__.side_effect = lambda: iter(handle.readlines.side_effect())
    return mo


def skip_if_platform(platform, hide=False):
    return skip_pred(lambda x: x.platform_name == platform,
                     'not supported for platform "{}"'.format(platform), hide)


def only_if_platform(platform, hide=False):
    return skip_pred(lambda x: x.platform_name != platform,
                     'only supported for platform "{}"'.format(platform), hide)


class AttrDict(object):
    def __init__(self, **kwargs):
        for k, v in iteritems(kwargs):
            setattr(self, k, v)


class CrossPlatformTestCase(TestCase):
    _platforms = ['linux', 'winnt', 'macos']

    def __init__(self, *args, **kwargs):
        clear_variables = kwargs.pop('clear_variables', False)
        variables = kwargs.pop('variables', {})
        self.platform_name = kwargs.pop('platform_name', None)

        TestCase.__init__(self, *args, **kwargs)
        if self.platform_name is None:
            return

        self.env = make_env(platform=self.platform_name,
                            clear_variables=clear_variables,
                            variables=variables)

    @property
    def Path(self):
        return self.env.host_platform.Path

    def shortDescription(self):
        return self.platform_name

    def parameterize(self):
        return parameterize_tests(self, platform_name=self._platforms)


class PathTestCase(TestCase):
    _path_infos = [
        (Path, os.path, 'native'),
        (PosixPath, posixpath, 'posix'),
        (WindowsPath, ntpath, 'windows'),
    ]

    def __init__(self, *args, **kwargs):
        info = kwargs.pop('path_info', None)

        TestCase.__init__(self, *args, **kwargs)
        if info is None:
            return
        self.Path, self.ospath, self._desc = info

    def shortDescription(self):
        return self._desc

    def parameterize(self):
        return parameterize_tests(self, path_info=self._path_infos)



from . import builder, cc, jvm
from .common import choose_builder
from ..languages import known_formats, known_langs

with known_langs.make('java') as x:
    x.vars(compiler='JAVAC', runner='JAVACMD', flags='JAVAFLAGS')
    x.exts(source=['.java'])

with known_langs.make('scala') as x:
    x.vars(compiler='SCALAC', runner='SCALACMD', flags='SCALAFLAGS')
    x.exts(source=['.scala'])

with known_formats.make('jvm', mode='dynamic') as x:
    x.vars(linker='JAR', flags='JARFLAGS')

_default_cmds = {
    'java' : ['javac', 'gcj'],
    'scala': 'scalac',
}

_builders = (jvm.JvmBuilder, cc.CcBuilder)


@builder('java', 'scala')
def java_builder(env, lang):
    return choose_builder(env, known_langs[lang], _default_cmds[lang],
                          _builders)



import json
import os
import platform
import sys
from collections import namedtuple
from six import string_types, iteritems

from . import platforms
from . import tools
from . import shell
from .backends import list_backends
from .file_types import Executable, Node
from .iterutils import first, isiterable, listify
from .path import InstallRoot, Path, Root
from .tools.common import Command
from .versioning import Version

LibraryMode = namedtuple('LibraryMode', ['shared', 'static'])


def try_to_json(value):
    return value.to_json() if value is not None else None


def try_from_json(type, value):
    return type.from_json(value) if value is not None else None


class EnvVersionError(RuntimeError):
    pass


class EnvVarDict(dict):
    def __setitem__(self, key, value):
        if ( not isinstance(key, string_types) or
             not isinstance(value, string_types) ):  # pragma: no cover
            raise TypeError('expected a string')
        dict.__setitem__(self, key, value)


class Toolchain(object):
    def __init__(self, path=None):
        self.path = path

    def to_json(self):
        return {
            'path': try_to_json(self.path),
        }

    @classmethod
    def from_json(cls, data):
        return cls(
            path=try_from_json(Path, data['path']),
        )


class Environment(object):
    version = 14
    envfile = '.bfg_environ'

    Mode = shell.Mode

    def __new__(cls, *args, **kwargs):
        env = object.__new__(cls)
        tools.init()
        env.__builders = {}
        env.__tools = {}
        return env

    def __init__(self, bfgdir, backend, backend_version, srcdir, builddir):
        self.bfgdir = bfgdir
        self.backend = backend
        self.backend_version = backend_version

        self.host_platform = platforms.host.platform_info()
        self.target_platform = platforms.target.platform_info()

        self.srcdir = srcdir
        self.builddir = builddir
        self.install_dirs = {}
        self.toolchain = Toolchain()

        self.initial_variables = dict(os.environ)
        self.init_variables()

    def finalize(self, install_dirs, library_mode, extra_args=None):
        # Fill in any install dirs that aren't already set (e.g. by a
        # toolchain file) with defaults from the target platform, but skip
        # absolute paths if this is a cross-compilation build.
        for k, v in iteritems(self.target_platform.install_dirs):
            if ( (self.is_cross and v and v.root == Root.absolute) or
                 self.install_dirs.get(k) ):
                continue
            self.install_dirs[k] = v

        for k, v in iteritems(install_dirs):
            if v:
                self.install_dirs[k] = v

        self.library_mode = LibraryMode(*library_mode)
        self.extra_args = extra_args

    def init_variables(self):
        self.variables = EnvVarDict(self.initial_variables)

    @property
    def is_cross(self):
        return self.host_platform != self.target_platform

    @property
    def base_dirs(self):
        dirs = {
            Root.srcdir: self.srcdir,
            Root.builddir: self.builddir
        }
        dirs.update(self.install_dirs)
        return dirs

    def getvar(self, key, default=None):
        return self.variables.get(key, default)

    def builder(self, lang):
        if lang not in self.__builders:
            self.__builders[lang] = tools.get_builder(self, lang)
        return self.__builders[lang]

    def tool(self, name):
        if name not in self.__tools:
            self.__tools[name] = tools.get_tool(self, name)
        return self.__tools[name]

    def _runner(self, lang):
        try:
            return self.builder(lang).runner
        except ValueError:
            try:
                return self.tool(tools.get_tool_runner(lang))
            except ValueError:
                return None

    def run_arguments(self, args, lang=None):
        if isinstance(args, Node):
            args = [args]
        elif isiterable(args):
            args = listify(args)
        else:
            return args

        if len(args) == 0 or not isinstance(args[0], Node):
            return args

        if lang is None:
            lang = first(getattr(args[0], 'lang', None), default=None)
        runner = self._runner(lang)
        if runner:
            return runner.run_arguments(args[0]) + args[1:]

        if not isinstance(args[0], Executable):
            raise TypeError('expected an executable for {} to run'
                            .format(lang))
        return args

    def execute(self, args, env=None, env_update=True, **kwargs):
        env_vars = self.variables
        if env:
            if env_update:
                env_vars = env_vars.copy()
                env_vars.update(env)
            else:
                env_vars = env

        if not kwargs.get('shell', False):
            args = Command.convert_args(args, lambda x: x.command)

        return shell.execute(args, env=env_vars, base_dirs=self.base_dirs,
                             **kwargs)

    def run(self, args, lang=None, *posargs, **kwargs):
        return self.execute(self.run_arguments(args, lang), *posargs, **kwargs)

    def save(self, path):
        with open(os.path.join(path, self.envfile), 'w') as out:
            json.dump({
                'version': self.version,
                'data': {
                    'bfgdir': self.bfgdir.to_json(),
                    'backend': self.backend,
                    'backend_version': str(self.backend_version),

                    'host_platform': self.host_platform.to_json(),
                    'target_platform': self.target_platform.to_json(),

                    'srcdir': self.srcdir.to_json(),
                    'builddir': self.builddir.to_json(),
                    'install_dirs': {
                        k.name: try_to_json(v)
                        for k, v in iteritems(self.install_dirs)
                    },
                    'toolchain': self.toolchain.to_json(),

                    'library_mode': self.library_mode,
                    'extra_args': self.extra_args,

                    'initial_variables': self.initial_variables,
                    'variables': self.variables,
                }
            }, out)

    @classmethod
    def load(cls, path):
        with open(os.path.join(path, cls.envfile)) as inp:
            state = json.load(inp)
            version, data = state['version'], state['data']
        if version > cls.version:
            raise EnvVersionError('saved version exceeds expected version')

        # Upgrade from older versions of the Environment if necessary.

        # v5 converts srcdir and builddir to Path objects internally.
        if version < 5:
            for i in ('srcdir', 'builddir'):
                data[i] = Path(data[i]).to_json()

        # v6 adds persistence for the backend's version and converts bfgpath to
        # a Path object internally.
        if version < 6:
            backend = list_backends()[data['backend']]
            data['backend_version'] = str(backend.version())
            data['bfgpath'] = Path(data['bfgpath']).to_json()

        # v7 replaces bfgpath with bfgdir.
        if version < 7:
            bfgdir = Path.from_json(data['bfgpath'] + (False,)).parent()
            data['bfgdir'] = bfgdir.to_json()
            del data['bfgpath']

        # v8 adds support for user-defined command-line arguments.
        if version < 8:
            data['extra_args'] = []

        # v9 adds options for choosing the mode to build libraries in.
        if version < 9:
            data['library_mode'] = [True, False]

        # v10 adds exec_prefix to install_dirs.
        if version < 10:
            data['install_dirs']['exec_prefix'] = ['', 'prefix']
            for i in ('bindir', 'libdir'):
                if data['install_dirs'][i][1] == 'prefix':
                    data['install_dirs'][i][1] = 'exec_prefix'

        # v11 adds $(DESTDIR) support to Path objects.
        if version < 11:
            for i in ('bfgdir', 'srcdir', 'builddir'):
                data[i] += (False,)
            for i in data['install_dirs']:
                data['install_dirs'][i] += (False,)

        # v12 splits platform into host_platform and target_platform.
        if version < 12:
            platform_name = data.pop('platform')
            data['host_platform'] = data['target_platform'] = platform_name

        # v13 adds initial_variables and toolchain.
        if version < 13:
            data['initial_variables'] = data['variables']
            data['toolchain'] = {'path': None}

        # v14 adds architecture to platform objects.
        if version < 14:
            for i in ('host_platform', 'target_platform'):
                genus, species = platforms.platform_tuple(data[i])
                data[i] = {'genus': genus, 'species': species,
                           'arch': platform.machine()}

        # Now that we've upgraded, initialize the Environment object.
        env = Environment.__new__(Environment)

        env.host_platform = platforms.host.from_json(data['host_platform'])
        env.target_platform = platforms.target.from_json(
            data['target_platform']
        )

        # With Python 2.x on Windows, the environment variables must all be
        # non-Unicode strings.
        if env.host_platform.family == 'windows' and sys.version_info[0] == 2:
            for key in ('initial_variables', 'variables'):
                data[key] = {str(k): str(v) for k, v in iteritems(data[key])}

        for i in ('backend', 'extra_args', 'initial_variables', 'variables'):
            setattr(env, i, data[i])

        for i in ('bfgdir', 'srcdir', 'builddir'):
            setattr(env, i, Path.from_json(data[i]))

        env.backend_version = Version(data['backend_version'])
        env.install_dirs = {
            InstallRoot[k]: Path.from_json(v) if v else None
            for k, v in iteritems(data['install_dirs'])
        }
        env.toolchain = Toolchain.from_json(data['toolchain'])
        env.library_mode = LibraryMode(*data['library_mode'])

        return env



import platform
import random

from .iterutils import iterate

try:  # pragma: no cover
    import winsound

    def _do_play(track):
        for freq, ms in track:
            winsound.Beep(freq, ms)
except ImportError:  # pragma: no cover
    from . import shell

    def _do_play(track):
        try:
            shell.which('beep')
            args = []
            for freq, ms in track:
                args.extend(['-n', '-f', freq, '-l', ms])
            args[0] = 'beep'
        except IOError:
            args = ['play', '-q']
            note = '|sox -n -p synth {} sawtooth {}'
            for freq, ms in track:
                args.append(note.format(ms / 1000.0, freq))
                args.append(note.format(ms / 8000.0, 0))
            args.extend(['vol', '-20dB'])

        try:
            shell.execute(args, stdout=shell.Mode.devnull,
                          stderr=shell.Mode.devnull)
        except OSError:
            raise ValueError("'sound player not found; install 'beep' or " +
                             "'sox' package?")


def riff(a, b, hold=3.5):
    def gen():
        for i, v in enumerate(b):
            yield a, 1
            yield a, 1
            d = 1 if i < len(b) - 1 else hold
            for j in iterate(v):
                yield j, d

    return list(gen())


E3 = 165
F3 = 175
FS3 = 185
G3 = 196
A3 = 220
B3 = 247
CS4 = 277
DS4 = 311
E4 = 330
F4 = 349
G4 = 392
GS4 = 415
A4 = 440
B4 = 494
CS5 = 554

e1m1_1 = riff(E3, (B3, A3, G3, F3, (FS3, G3), B3, A3, G3, F3))
e1m1_2 = riff(A3, (A4, G4, F4, DS4, (E4, F4), A4, G4, F4, DS4))
e1m1_3 = (riff(CS4, (CS5, B4, A4, G4, (GS4, A4)), hold=1) +
          riff(B3, (B4, A4, G4), hold=1) + riff(A3, (F4,)))

e1m1 = e1m1_1 * 2 + e1m1_2 + e1m1_1 + e1m1_3 + e1m1_1

plat = platform.system()
msgs = [
    "Please don't leave, there's more demons to toast!",
    "Let's beat it -- This is turning into a bloodbath!",
    "I wouldn't leave if I were you. {} is much worse.".format(plat),
    "You're trying to say you like {} better than me, right?".format(plat),
    "Don't leave yet -- There's a demon around that corner!",
    "Ya know, next time you come in here I'm gonna toast ya.",
    'Go ahead and leave. See if I care.',
    'Are you sure you want to quit this great music?',
]


def play(tempo, long):
    try:
        _do_play((freq, int(dur * 15000 / tempo))
                 for freq, dur in (e1m1 if long else e1m1_1))
    except KeyboardInterrupt:  # pragma: no cover
        print(random.choice(msgs))
        exit(1)



from collections import defaultdict

from . import builtin
from .. import options as opts
from .file_types import FileList, static_file
from ..backends.make import writer as make
from ..backends.ninja import writer as ninja
from ..build_inputs import build_input, Edge
from ..file_types import *
from ..iterutils import first, iterate
from ..languages import known_langs
from ..objutils import convert_each, convert_one
from ..path import Path
from ..shell import posix as pshell

build_input('compile_options')(lambda build_inputs, env: defaultdict(list))


class BaseCompile(Edge):
    desc_verb = 'compile'

    def __init__(self, build, name, internal_options, directory=None,
                 extra_deps=None, description=None):
        if name is None:
            name = self.compiler.default_name(self.file, self)
            if directory:
                name = directory.append(name).suffix

        extra_options = self.compiler.pre_build(build, name, self)
        output = self.compiler.output_file(name, self)
        primary = first(output)

        self._internal_options = opts.option_list(
            internal_options, extra_options
        )

        options = self.options
        public_output = self.compiler.post_build(build, options, output, self)
        primary.post_install = self.compiler.post_install(options, output,
                                                          self)

        Edge.__init__(self, build, output, public_output, extra_deps,
                      description)

    @property
    def options(self):
        return self._internal_options + self.user_options

    @property
    def flags(self):
        return self.compiler.flags(self.options, self.raw_output)

    @staticmethod
    def _convert_args_lang(kwargs):
        lang = kwargs.get('lang')
        src_lang = known_langs[lang].src_lang if lang else None
        return lang, src_lang

    @staticmethod
    def convert_args(kwargs):
        convert_one(kwargs, 'directory', Path.ensure, strict=True)
        return kwargs


class Compile(BaseCompile):
    def __init__(self, build, name, includes, include_deps, pch, libs,
                 packages, options, lang=None, directory=None, extra_deps=None,
                 description=None):
        self.includes = includes
        self.include_deps = include_deps
        self.packages = packages
        self.user_options = options

        internal_options = opts.option_list(
            (i.compile_options(self.compiler) for i in self.packages),
            (opts.include_dir(i) for i in self.includes)
        )

        self.pch = pch
        if self.pch:
            if not self.compiler.accepts_pch:
                raise TypeError('pch not supported for this compiler')
            internal_options.append(opts.pch(self.pch))

        # Don't bother handling forward_opts from libs now, since the only
        # languages that need libs during compilation don't support static
        # linking anyway.
        if self.compiler.needs_libs:
            self.libs = libs
            internal_options.extend(opts.lib(i) for i in self.libs)

        BaseCompile.__init__(self, build, name, internal_options, directory,
                             extra_deps, description)

    @staticmethod
    def convert_args(builtins, lang, src_lang, kwargs):
        def pch(file, **kwargs):
            return builtins['precompiled_header'](file, file, **kwargs)

        includes = kwargs.get('includes')
        kwargs['include_deps'] = [
            i for i in iterate(includes)
            if isinstance(i, SourceCodeFile) or getattr(i, 'creator', None)
        ]
        convert_each(kwargs, 'includes', builtins['header_directory'])

        convert_each(kwargs, 'libs', builtins['library'], lang=src_lang)
        convert_each(kwargs, 'packages', builtins['package'], lang=src_lang)

        kwargs['options'] = pshell.listify(kwargs.get('options'),
                                           type=opts.option_list)

        convert_one(kwargs, 'pch', pch, includes=includes,
                    packages=kwargs['packages'], options=kwargs['options'],
                    lang=lang)

        kwargs = BaseCompile.convert_args(kwargs)
        return kwargs

    def add_extra_options(self, options):
        self._internal_options.extend(options)
        # PCH files should always be built with the same options as files using
        # them, so forward the extra options onto the PCH if it exists.
        if self.pch and hasattr(self.pch.creator, 'add_extra_options'):
            self.pch.creator.add_extra_options(options)


class CompileSource(Compile):
    def __init__(self, build, env, name, file, lang=None, **kwargs):
        self.file = file
        if self.file.lang is None:
            raise ValueError('unable to determine language for file {!r}'
                             .format(self.file.path))

        self.compiler = env.builder(lang or self.file.lang).compiler
        Compile.__init__(self, build, name, **kwargs)

    @classmethod
    def convert_args(cls, builtins, file, kwargs):
        lang, src_lang = cls._convert_args_lang(kwargs)
        file = builtins['source_file'](file, lang=src_lang)
        return file, Compile.convert_args(builtins, lang, src_lang, kwargs)


class CompileHeader(Compile):
    desc_verb = 'compile-header'

    def __init__(self, build, env, name, file, source, lang=None, **kwargs):
        self.file = file
        if self.file.lang is None:
            raise ValueError('unable to determine language for file {!r}'
                             .format(self.file.path))

        self.pch_source = source

        self.compiler = env.builder(lang or self.file.lang).pch_compiler
        Compile.__init__(self, build, name, **kwargs)

    @classmethod
    def convert_args(cls, builtins, file, kwargs):
        lang, src_lang = cls._convert_args_lang(kwargs)
        file = builtins['header_file'](file, lang=src_lang)
        convert_one(kwargs, 'source', builtins['source_file'], lang=file.lang)
        return file, Compile.convert_args(builtins, lang, src_lang, kwargs)


class GenerateSource(BaseCompile):
    desc_verb = 'generate'

    def __init__(self, build, env, name, file, options, lang=None,
                 directory=None, extra_deps=None, description=None):
        self.file = file
        if not isinstance(self.file, CodeFile):
            raise ValueError('unable to determine language for file {!r}'
                             .format(self.file.path))

        self.user_options = options

        self.compiler = env.builder(lang or self.file.lang).transpiler
        BaseCompile.__init__(self, build, name, None, directory, extra_deps,
                             description)

    @classmethod
    def convert_args(cls, builtins, file, kwargs):
        lang, src_lang = cls._convert_args_lang(kwargs)
        file = builtins['auto_file'](file, lang=src_lang)

        kwargs['options'] = pshell.listify(kwargs.get('options'),
                                           type=opts.option_list)
        kwargs = BaseCompile.convert_args(kwargs)
        return file, kwargs


@builtin.function('builtins', 'build_inputs', 'env')
@builtin.type(ObjectFile, extra_in_type=type(None))
def object_file(builtins, build, env, name=None, file=None, **kwargs):
    if file is None:
        if name is None:
            raise TypeError('expected name')
        dist = kwargs.pop('dist', True)
        params = [('format', env.target_platform.object_format),
                  ('lang', build['project']['lang'])]
        return static_file(build, ObjectFile, name, dist, params, kwargs)
    file, kwargs = CompileSource.convert_args(builtins, file, kwargs)
    return CompileSource(build, env, name, file, **kwargs).public_output


@builtin.function('builtins', 'build_inputs', 'env')
@builtin.type(FileList, in_type=object)
def object_files(builtins, build, env, files, **kwargs):
    @builtin.type(ObjectFile, extra_in_type=SourceFile)
    def make_object_file(file, **kwargs):
        file, kwargs = CompileSource.convert_args(builtins, file, kwargs)
        return CompileSource(build, env, None, file, **kwargs).public_output

    return FileList(make_object_file, files, **kwargs)


@builtin.function('builtins', 'build_inputs', 'env')
@builtin.type(PrecompiledHeader, extra_in_type=type(None))
def precompiled_header(builtins, build, env, name=None, file=None, **kwargs):
    if file is None:
        if name is None:
            raise TypeError('expected name')
        dist = kwargs.pop('dist', True)
        params = [('lang', build['project']['lang'])]
        return static_file(build, PrecompiledHeader, name, dist, params,
                           kwargs)
    file, kwargs = CompileHeader.convert_args(builtins, file, kwargs)
    return CompileHeader(build, env, name, file, **kwargs).public_output


@builtin.function('builtins', 'build_inputs', 'env')
@builtin.type(SourceCodeFile, short_circuit=False, first_optional=True)
def generated_source(builtins, build, env, name, file, **kwargs):
    file, kwargs = GenerateSource.convert_args(builtins, file, kwargs)
    return GenerateSource(build, env, name, file, **kwargs).public_output


@builtin.function('builtins')
@builtin.type(FileList, in_type=object)
def generated_sources(builtins, files, **kwargs):
    return FileList(builtins['generated_source'], files, **kwargs)


@builtin.function('build_inputs')
def global_options(build, options, lang):
    for i in iterate(lang):
        build['compile_options'][i].extend(pshell.listify(
            options, type=opts.option_list
        ))


def _get_flags(backend, rule, build_inputs, buildfile):
    variables = {}
    cmd_kwargs = {}

    compiler = rule.compiler
    if hasattr(compiler, 'flags_var'):
        global_cflags, cflags = backend.flags_vars(
            compiler.flags_var,
            ( compiler.global_flags +
              compiler.flags(build_inputs['compile_options'][compiler.lang]) ),
            buildfile
        )
        cmd_kwargs['flags'] = cflags
        flags = rule.flags
        if flags:
            variables[cflags] = [global_cflags] + flags

    return variables, cmd_kwargs


@make.rule_handler(CompileSource, CompileHeader, GenerateSource)
def make_compile(rule, build_inputs, buildfile, env):
    compiler = rule.compiler
    variables, cmd_kwargs = _get_flags(make, rule, build_inputs, buildfile)

    output_params = []
    if compiler.num_outputs == 'all':
        output_vars = make.qvar('@')
    else:
        output_vars = []
        for i in range(compiler.num_outputs):
            v = make.var(str(i + 1))
            output_vars.append(v)
            output_params.append(rule.output[i])

    recipename = make.var('RULE_{}'.format(compiler.rule_name.upper()))
    if not buildfile.has_variable(recipename):
        recipe_extra = []

        # Only GCC-style depfiles are supported by Make.
        if compiler.deps_flavor == 'gcc':
            depfixer = env.tool('depfixer')
            cmd_kwargs['deps'] = deps = first(output_vars) + '.d'
            recipe_extra = [make.Silent(depfixer(deps))]

            depfile = rule.output[0].path.addext('.d')
            build_inputs.add_target(File(depfile))
            buildfile.include(depfile, optional=True)

        buildfile.define(recipename, [compiler(
            make.qvar('<'), output_vars, **cmd_kwargs
        )] + recipe_extra)

    deps = []
    if getattr(rule, 'pch_source', None):
        deps.append(rule.pch_source)
    deps.append(rule.file)
    if getattr(rule, 'pch', None):
        deps.append(rule.pch)
    deps.extend(getattr(rule, 'include_deps', []))
    if compiler.needs_libs:
        deps.extend(rule.libs)

    make.multitarget_rule(
        buildfile,
        targets=rule.output,
        deps=deps + rule.extra_deps,
        order_only=make.directory_deps(rule.output),
        recipe=make.Call(recipename, *output_params),
        variables=variables
    )


@ninja.rule_handler(CompileSource, CompileHeader, GenerateSource)
def ninja_compile(rule, build_inputs, buildfile, env):
    compiler = rule.compiler
    variables, cmd_kwargs = _get_flags(ninja, rule, build_inputs, buildfile)
    if rule.description:
        variables['description'] = rule.description

    if compiler.num_outputs == 'all':
        output_vars = ninja.var('out')
    elif compiler.num_outputs == 1:
        output_vars = ninja.var('output')
        variables[output_vars] = rule.output[0]
    else:
        output_vars = []
        for i in range(compiler.num_outputs):
            v = ninja.var('output{}'.format(i + 1))
            output_vars.append(v)
            variables[v] = rule.output[i]

    if not buildfile.has_rule(compiler.rule_name):
        depfile = None
        deps = None

        if compiler.deps_flavor == 'gcc':
            deps = 'gcc'
            cmd_kwargs['deps'] = depfile = ninja.var('out') + '.d'
        elif compiler.deps_flavor == 'msvc':
            deps = 'msvc'
            cmd_kwargs['deps'] = True

        desc = rule.desc_verb + ' => ' + first(output_vars)
        buildfile.rule(name=compiler.rule_name, command=compiler(
            ninja.var('in'), output_vars, **cmd_kwargs
        ), depfile=depfile, deps=deps, description=desc)

    inputs = [rule.file]
    implicit_deps = []
    if getattr(rule, 'pch', None):
        implicit_deps.append(rule.pch)
    if getattr(rule, 'pch_source', None):
        inputs = [rule.pch_source]
        implicit_deps.append(rule.file)
    implicit_deps.extend(getattr(rule, 'include_deps', []))
    if compiler.needs_libs:
        implicit_deps.extend(rule.libs)

    # Ninja doesn't support multiple outputs and deps-parsing at the same time,
    # so just use the first output and set up an alias if necessary. Aliases
    # aren't perfect, since the build can get out of sync if you delete the
    # "alias" file, but it's close enough.
    if compiler.deps_flavor in ('gcc', 'msvc') and len(rule.output) > 1:
        output = rule.output[0]
        buildfile.build(
            output=rule.output[1:],
            rule='phony',
            inputs=rule.output[0]
        )
    else:
        output = rule.output

    buildfile.build(
        output=output,
        rule=compiler.rule_name,
        inputs=inputs,
        implicit=implicit_deps + rule.extra_deps,
        variables=variables
    )


try:
    from ..backends.msbuild import writer as msbuild

    @msbuild.rule_handler(CompileSource, CompileHeader)
    def msbuild_compile(rule, build_inputs, solution, env):
        # MSBuild does compilation and linking in one unit; see link.py.
        pass
except ImportError:  # pragma: no cover
    pass



import string as _string
from six.moves import filter as ifilter
from six import string_types
from six.moves import zip

from . import iterutils


class safe_string(object):
    def _safe_format(self, format_spec):
        return self

    def __add__(self, rhs):
        return jbos(self, safe_str(rhs))

    def __radd__(self, lhs):
        return jbos(safe_str(lhs), self)


class safe_string_ops(object):
    def __add__(self, rhs):
        return jbos(safe_str(self), safe_str(rhs))

    def __radd__(self, lhs):
        return jbos(safe_str(lhs), safe_str(self))


stringy_types = string_types + (safe_string,)


def safe_str(s):
    if isinstance(s, stringy_types):
        return s
    elif hasattr(s, '_safe_str'):
        return s._safe_str()
    else:
        raise NotImplementedError(type(s))


class literal_types(safe_string):
    def __init__(self, string):
        if not isinstance(string, string_types):
            raise TypeError('expected a string')
        self.string = string

    def __str__(self):
        raise NotImplementedError()

    def __repr__(self):
        return '`{}`'.format(self.string)

    def __eq__(self, rhs):
        if type(self) is not type(rhs):
            return NotImplemented
        return self.string == rhs.string

    def __ne__(self, rhs):
        return not (self == rhs)


class shell_literal(literal_types):
    """A string which has already been escaped for shell purposes, useful if
    you want to use characters with syntactic meaning in your shell (e.g. to
    use I/O redirection."""


class literal(literal_types):
    """A string which has already been escaped for *all* purposes (read: both
    shell and build files), useful if you want to use characters with syntactic
    meaning in your build script."""


class jbos(safe_string):  # Just a Bunch of Strings
    def __init__(self, *args):
        self.__bits = tuple(self.__canonicalize(args))

    @staticmethod
    def __canonicalize(value):
        def flatten_bits(value):
            for i in value:
                if isinstance(i, jbos):
                    for j in i.bits:
                        yield j
                elif isinstance(i, stringy_types):
                    yield i
                else:
                    raise TypeError(type(i))

        bits = ifilter(None, flatten_bits(value))
        try:
            last = next(bits)
        except StopIteration:
            return

        for i in bits:
            same_type = type(i) == type(last)
            if same_type and isinstance(i, string_types):
                last += i
            elif same_type and isinstance(i, literal_types):
                last = type(i)(last.string + i.string)
            else:
                yield last
                last = i
        yield last

    @property
    def bits(self):
        return self.__bits

    def simplify(self):
        if len(self.bits) == 0:
            return ''
        elif len(self.bits) == 1:
            return self.bits[0]
        return self

    def __str__(self):
        raise NotImplementedError()

    def __repr__(self):
        return '|{}|'.format(', '.join(repr(i) for i in self.bits))

    def __eq__(self, rhs):
        if type(self) is not type(rhs):
            return NotImplemented
        return ( len(self.bits) == len(rhs.bits) and
                 all(i == j for i, j in zip(self.bits, rhs.bits)) )

    def __ne__(self, rhs):
        return not (self == rhs)


def join(iterable, delim):
    if delim:
        iterable = iterutils.tween(iterable, delim)
    return sum(iterable, jbos()).simplify()


def format_field(value, format_spec):
    t = type(value)
    if hasattr(t, '_safe_format'):
        return t._safe_format(value, format_spec)
    elif hasattr(value, '_safe_str'):
        return format_field(value._safe_str(), format_spec)
    return __builtins__['format'](value, format_spec)


class Formatter(_string.Formatter):
    def format_field(self, value, format_spec):
        return format_field(value, format_spec)

    # We need to copy these from the Python stdlib implementation to change the
    # join() call at the end of _vformat (and to get automatic field numbering
    # in Python 2.7...)
    def vformat(self, format_string, args, kwargs):
        used_args = set()
        result, _ = self._vformat(format_string, args, kwargs, used_args, 2)
        self.check_unused_args(used_args, args, kwargs)
        return result

    def _vformat(self, format_string, args, kwargs, used_args, recursion_depth,
                 auto_arg_index=0):
        if recursion_depth < 0:  # pragma: no cover
            raise ValueError('Max string recursion exceeded')
        result = []
        for literal_text, field_name, format_spec, conversion in \
                self.parse(format_string):

            # output the literal text
            if literal_text:
                result.append(literal_text)

            # if there's a field, output it
            if field_name is not None:
                # this is some markup, find the object and do
                #  the formatting

                # handle arg indexing when empty field_names are given.
                if field_name == '':
                    if auto_arg_index is False:
                        raise ValueError('cannot switch from manual field '
                                         'specification to automatic field '
                                         'numbering')
                    field_name = str(auto_arg_index)
                    auto_arg_index += 1
                elif field_name.isdigit():
                    if auto_arg_index:
                        raise ValueError('cannot switch from manual field '
                                         'specification to automatic field '
                                         'numbering')
                    # disable auto arg incrementing, if it gets
                    # used later on, then an exception will be raised
                    auto_arg_index = False

                # given the field_name, find the object it references
                #  and the argument it came from
                obj, arg_used = self.get_field(field_name, args, kwargs)
                used_args.add(arg_used)

                # do any conversion on the resulting object
                obj = self.convert_field(obj, conversion)

                # expand the format spec, if needed
                format_spec, auto_arg_index = self._vformat(
                    format_spec, args, kwargs,
                    used_args, recursion_depth - 1,
                    auto_arg_index=auto_arg_index)

                # format the object and append to the result
                result.append(self.format_field(obj, format_spec))

        return join(result, ''), auto_arg_index


_fmt = Formatter()


def safe_format(*args, **kwargs):
    return _fmt.format(*args, **kwargs)



import re as _re
from argparse import *

from .. import path as _path

_ArgumentParser = ArgumentParser


class ToggleAction(Action):
    def __init__(self, option_strings, dest, default=False, required=False,
                 help=None):
        if len(option_strings) == 0:
            raise ValueError('option string must begin with "--"')

        self.true_strings = [self._prefix(i, self._true_prefix)
                             for i in option_strings]
        self.false_strings = [self._prefix(i, self._false_prefix)
                              for i in option_strings]

        option_strings = self.true_strings + self.false_strings
        Action.__init__(self, option_strings, dest=dest, nargs=0,
                        default=default, required=required, help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        value = option_string in self.true_strings
        setattr(namespace, self.dest, value)

    @staticmethod
    def _prefix(s, prefix):
        if not s.startswith('--'):
            raise ValueError('option string must begin with "--"')
        return _re.sub('(^--(x-)?)', r'\1' + prefix, s)


class EnableAction(ToggleAction):
    _true_prefix = 'enable-'
    _false_prefix = 'disable-'


class WithAction(ToggleAction):
    _true_prefix = 'with-'
    _false_prefix = 'without-'


class ArgumentParser(_ArgumentParser):
    def __init__(self, *args, **kwargs):
        _ArgumentParser.__init__(self, *args, **kwargs)
        self.register('action', 'enable', EnableAction)
        self.register('action', 'with', WithAction)

    def _get_option_tuples(self, option_string):
        # Don't try to check prefixes for long options; this is similar to
        # Python 3.5's `allow_abbrev=False`, except this doesn't break combined
        # short options. See <https://bugs.python.org/issue26967>.
        if option_string[:2] == self.prefix_chars * 2:
            return []

        return _ArgumentParser._get_option_tuples(self, option_string)


class BaseFile(object):
    def __init__(self, check_type, kind, must_exist=False):
        self._check_type = check_type
        self._kind = kind
        self.must_exist = must_exist

    def __call__(self, string):
        p = _path.abspath(string)
        if _path.exists(p):
            if not self._check_type(p):
                raise ArgumentTypeError("'{}' is not a {}"
                                        .format(string, self._kind))
        elif self.must_exist:
            raise ArgumentTypeError("'{}' does not exist".format(string))

        return p


class Directory(BaseFile):
    def __init__(self, *args, **kwargs):
        BaseFile.__init__(self, _path.isdir, 'directory', *args, **kwargs)


class File(BaseFile):
    def __init__(self, *args, **kwargs):
        BaseFile.__init__(self, _path.isfile, 'file', *args, **kwargs)


# It'd be nice to just have a UserArgumentParser class with this method but it
# wouldn't propagate to argument groups, so it's easier to just do it this way.
def add_user_argument(parser, *names, **kwargs):
    if any(not i.startswith('--') for i in names):
        raise ValueError('option string must begin with "--"')
    if any(i.startswith('--x-') for i in names):
        raise ValueError('"x-" prefix is reserved')

    if parser.usage == 'parse':
        names += tuple('--x-' + i[2:] for i in names)
    return parser.add_argument(*names, **kwargs)



from . import builtin
from .. import versioning as v
from ..objutils import objectify


@builtin.function(context='*')
def bfg9000_required_version(version=None, python_version=None):
    version = objectify(version or '', v.PythonSpecifierSet, prereleases=True)
    python_version = objectify(python_version or '', v.PythonSpecifierSet,
                               prereleases=True)

    v.check_version(v.bfg_version, version, kind='bfg9000')
    v.check_version(v.python_version, python_version, kind='python')


@builtin.getter(context='*')
def bfg9000_version():
    return v.bfg_version



from itertools import chain
from six import iteritems, iterkeys

from .iterutils import listify


def _get_prop(attr, desc):
    def inner(self, key):
        try:
            return getattr(self, attr)[key]
        except KeyError:
            raise ValueError('{} {!r} does not support {} {!r}'
                             .format(self._kind, self.name, desc, key))
    return inner


class _Info(object):
    _fields = ('vars',)

    def __init__(self, name, vars):
        self.name = name
        self._var = vars

    var = _get_prop('_var', 'var')


class _LanguageInfo(_Info):
    _kind = 'language'
    _fields = _Info._fields + ('exts', 'auxexts')

    def __init__(self, name, base, vars, exts, auxexts):
        _Info.__init__(self, name, vars)
        self._base = base

        allkeys = set(iterkeys(exts)) | set(iterkeys(auxexts))
        self._exts = {i: listify(exts.get(i)) for i in allkeys}
        self._auxexts = {i: listify(auxexts.get(i)) for i in allkeys}

    exts = _get_prop('_exts', 'file type')
    auxexts = _get_prop('_auxexts', 'file type')

    @property
    def src_lang(self):
        return self._base or self.name

    def allexts(self, key):
        return self.exts(key) + self.auxexts(key)

    def default_ext(self, key):
        exts = self.exts(key)
        if len(exts):
            return exts[0]
        return self.auxexts(key)[0]

    def extkind(self, ext):
        for k, v in chain(iteritems(self._exts), iteritems(self._auxexts)):
            if ext in v:
                return k
        return None


class _FormatInfo(_Info):
    _kind = 'format'


class _Definer(object):
    def __init__(self, parent, type, name, **kwargs):
        self._parent = parent
        self._type = type
        self._name = name
        self._fields = set(type._fields)

        self._data = {i: {} for i in type._fields}
        for k, v in iteritems(kwargs):
            assert k not in self._fields
            self._data[k] = v

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self._parent._add(self._type(self._name, **self._data))

    def __getattr__(self, attr):
        if attr not in self._fields:
            raise AttributeError(attr)

        def inner(**kwargs):
            self._data[attr] = kwargs

        return inner


class Languages(object):
    def __init__(self):
        self._langs = {}
        self._ext2lang = {}

    def __getitem__(self, name):
        try:
            return self._langs[name]
        except KeyError:
            raise ValueError('unrecognized language {!r}'.format(name))

    def __contains__(self, name):
        return name in self._langs

    def _add(self, info):
        self._langs[info.name] = info
        for kind, exts in iteritems(info._exts):
            for ext in exts:
                if ext in self._ext2lang:
                    raise ValueError('{ext!r} already used by {lang!r}'.format(
                        ext=ext, lang=self._ext2lang[ext][0]
                    ))
                self._ext2lang[ext] = (info.name, kind)

    def fromext(self, ext, kind):
        lang, langkind = self.extinfo(ext)
        return lang if langkind == kind else None

    def extinfo(self, ext):
        return self._ext2lang.get(ext, (None, None))

    def make(self, name, base=None):
        return _Definer(self, _LanguageInfo, name, base=base)


class Formats(object):
    def __init__(self):
        self._formats = {}

    def __getitem__(self, name):
        try:
            return self._formats[name[0]][name[1]]
        except KeyError:
            raise ValueError("unrecognized format '{} ({})'"
                             .format(name[0], name[1]))

    def _add(self, info):
        name, mode = info.name
        if name not in self._formats:
            self._formats[name] = {}
        self._formats[name][mode] = info

    def make(self, name, mode):
        return _Definer(self, _FormatInfo, (name, mode))


known_langs = Languages()
known_formats = Formats()



import tarfile

from . import *


class TestFindFiles(IntegrationTest):
    def __init__(self, *args, **kwargs):
        IntegrationTest.__init__(
            self, os.path.join(examples_dir, '06_find_files'), *args, **kwargs
        )

    def test_hello(self):
        self.build(executable('hello'))
        self.assertOutput([executable('hello')], 'Hello, world!\n')

    def test_goodbye(self):
        self.build(executable('goodbye'))
        self.assertOutput([executable('goodbye')],
                          'Goodbye!\nAuf Wiedersehen!\n')

    @skip_if_backend('msbuild')
    def test_dist(self):
        dist = output_file('06_find_files.tar.gz')
        self.build('dist')
        self.assertExists(dist)
        with tarfile.open(self.target_path(dist)) as t:
            self.assertEqual(set(t.getnames()), {
                '06_find_files/build.bfg',
                '06_find_files/src/hello/hello.cpp',
                '06_find_files/src/hello/hello.hpp',
                '06_find_files/src/hello/main.cpp',
                '06_find_files/src/goodbye',
                '06_find_files/src/goodbye/main.cpp',
                '06_find_files/src/goodbye/english',
                '06_find_files/src/goodbye/english/goodbye.cpp',
                '06_find_files/src/goodbye/english/goodbye.hpp',
                '06_find_files/src/goodbye/german',
                '06_find_files/src/goodbye/german/goodbye.cpp',
                '06_find_files/src/goodbye/german/goodbye.hpp',
            })



from . import builder, cc
from .common import choose_builder
from ..languages import known_langs

with known_langs.make('f77') as x:
    x.vars(compiler='FC', flags='FFLAGS')
    x.exts(source=['.f', '.for', '.ftn'])

with known_langs.make('f95') as x:
    x.vars(compiler='FC', flags='FFLAGS')
    x.exts(source=['.f90', '.f95', '.f03', '.f08'])

_default_cmds = ['gfortran']
_builders = (cc.CcBuilder,)


@builder('f77', 'f95')
def fortran_builder(env, lang):
    return choose_builder(env, known_langs[lang], _default_cmds, _builders)



from contextlib import contextmanager
from six import string_types

from . import builtin
from .find import exclude_globs
from ..file_types import *
from ..iterutils import iterate, uniques
from ..languages import known_langs
from ..path import Path, Root, makedirs as _makedirs

_kind_to_file_type = {
    'header': HeaderFile,
    'source': SourceFile,
    'resource': ResourceFile,
}


def static_file(build, file_type, name, dist=True, params=[], kwargs={}):
    extra_args = []
    for key, default in params:
        extra_args.append(kwargs.pop(key, default))
    if kwargs:
        raise TypeError("unexpected keyword argument '{}'".format(
            next(iter(kwargs))
        ))

    path = Path.ensure(name, Root.srcdir)
    file = file_type(path, *extra_args)
    if dist and path.root == Root.srcdir:
        build.add_source(file)
    return file


class FileList(list):
    def __init__(self, fn, files, **kwargs):
        list.__init__(self, (fn(i, **kwargs) for i in iterate(files)))

    def __getitem__(self, key):
        if isinstance(key, string_types):
            key = Path(key, Root.srcdir)
        elif isinstance(key, File):
            key = key.path

        if isinstance(key, Path):
            for i in self:
                if i.creator and i.creator.file.path == key:
                    return i
            raise IndexError('{!r} not found'.format(key))
        else:
            return list.__getitem__(self, key)


@contextmanager
def make_immediate_file(build, env, file, mode='w', makedirs=True):
    if makedirs:
        _makedirs(file.path.parent().string(env.base_dirs), exist_ok=True)

    yield open(file.path.string(env.base_dirs), mode)
    build['regenerate'].outputs.append(file)


@builtin.function('build_inputs')
@builtin.type(File)
def generic_file(build, name, dist=True):
    return static_file(build, File, name, dist)


@builtin.function('build_inputs')
@builtin.type(SourceFile)
def source_file(build, name, lang=None, dist=True):
    path = Path.ensure(name, Root.srcdir)
    lang = lang or known_langs.fromext(path.ext(), 'source')
    return static_file(build, SourceFile, path, dist, [('lang', lang)])


@builtin.function('build_inputs')
@builtin.type(ResourceFile)
def resource_file(build, name, lang=None, dist=True):
    path = Path.ensure(name, Root.srcdir)
    lang = lang or known_langs.fromext(path.ext(), 'resource')
    return static_file(build, ResourceFile, path, dist, [('lang', lang)])


@builtin.function('build_inputs')
@builtin.type(HeaderFile)
def header_file(build, name, lang=None, dist=True):
    path = Path.ensure(name, Root.srcdir)
    lang = lang or known_langs.fromext(path.ext(), 'header')
    return static_file(build, HeaderFile, path, dist, [('lang', lang)])


@builtin.function('build_inputs')
@builtin.type(ModuleDefFile)
def module_def_file(build, name, dist=True):
    return static_file(build, ModuleDefFile, name, dist)


@builtin.function('build_inputs')
@builtin.type(File)
def auto_file(build, name, lang=None, dist=True):
    path = Path.ensure(name, Root.srcdir)
    if lang:
        kind = None
        if lang in known_langs:
            kind = known_langs[lang].extkind(path.ext())
    else:
        lang, kind = known_langs.extinfo(path.ext())

    if lang:
        return static_file(build, _kind_to_file_type[kind or 'source'], path,
                           dist, [('lang', lang)])
    return static_file(build, File, path, dist)


# These builtins will find all the files in a directory so that they can be
# added to the distribution. XXX: Perhaps these could be reworked so that
# adding/removing files in directories doesn't force bfg to regenerate build
# files.


def _find(builtins, path, include, type, extra, *args, **kwargs):
    if not include and not extra:
        return None
    return builtins['find_files'](path, include, type, extra, *args, **kwargs)


def _directory_path(thing):
    if isinstance(thing, File):
        return thing.path.parent()
    else:
        return Path.ensure(thing, Root.srcdir)


@builtin.function('builtins', 'build_inputs')
@builtin.type(Directory, extra_in_type=File)
def directory(builtins, build, name, include=None, extra=None,
              exclude=exclude_globs, filter=None, dist=True, cache=True):
    path = _directory_path(name)
    files = _find(builtins, path, include, '*', extra, exclude, filter,
                  dist=dist, cache=cache)
    return static_file(build, Directory, path, dist, [('files', files)])


@builtin.function('builtins', 'build_inputs')
@builtin.type(HeaderDirectory, extra_in_type=SourceCodeFile)
def header_directory(builtins, build, name, include=None, extra=None,
                     exclude=exclude_globs, filter=None, system=False,
                     lang=None, dist=True, cache=True):
    def header_file(*args, **kwargs):
        return builtins['header_file'](*args, lang=lang, **kwargs)

    if isinstance(name, SourceCodeFile):
        lang = name.lang

    path = _directory_path(name)
    files = _find(builtins, path, include, 'f', extra, exclude, filter,
                  file_type=header_file, dist=dist, cache=cache)
    langs = uniques(i.lang for i in files if i.lang) if files else lang

    params = [('files', files), ('system', system), ('langs', langs)]
    return static_file(build, HeaderDirectory, path, dist, params)



from . import safe_str as _safe_str
from .iterutils import listify as _listify
from .path import InstallRoot as _InstallRoot, install_path as _install_path


def file_install_path(file, cross=None):
    return _install_path(file.path, file.install_root,
                         directory=isinstance(file, Directory), cross=cross)


def installify(file, cross=None):
    return file.clone(lambda f: file_install_path(f, cross), True)


def _clone_traits(exclude=set(), subfiles={}):
    def inner(cls):
        cls._clone_exclude = cls._clone_exclude | exclude
        if subfiles:
            cls._clone_subfiles = cls._clone_subfiles.copy()
            cls._clone_subfiles.update(subfiles)
        return cls

    return inner


class Node(_safe_str.safe_string_ops):
    private = False

    def __init__(self, path):
        self.creator = None
        self.path = path

    def _safe_str(self):
        return _safe_str.safe_str(self.path)

    @property
    def all(self):
        return [self]

    def __repr__(self):
        return '<{type} {name}>'.format(
            type=type(self).__name__, name=repr(self.path)
        )

    def __hash__(self):
        return hash(self.path)

    def __eq__(self, rhs):
        return type(self) == type(rhs) and self.path == rhs.path

    def __ne__(self, rhs):
        return not (self == rhs)


class Phony(Node):
    pass


class File(Node):
    _clone_exclude = {'path', 'creator', 'post_install'}
    _clone_subfiles = {}

    install_kind = None
    install_root = None

    def __init__(self, path):
        Node.__init__(self, path)
        self.post_install = None

    @property
    def install_deps(self):
        return []

    def _clone_args(self, pathfn, recursive):
        args = {'path': pathfn(self)}
        for k, v in self.__dict__.items():
            if k in self._clone_exclude:
                continue
            try:
                dest = self._clone_subfiles[k]
                orig = getattr(self, k)
                if orig is None:
                    args[dest] = None
                elif recursive:
                    args[dest] = pathfn(orig)
                else:
                    args[dest] = orig.path
            except KeyError:
                args[k] = v
        return args

    def clone(self, pathfn, recursive=False, inner=None):
        clone = type(self)(**self._clone_args(pathfn, recursive))
        if inner and inner is not self:
            for i in self._clone_subfiles:
                if getattr(self, i) is inner:
                    return getattr(clone, i)
            raise RuntimeError('unable to find inner clone object')
        return clone


@_clone_traits(exclude={'files'})
class Directory(File):
    def __init__(self, path, files=None):
        File.__init__(self, path)
        self.files = files

    def _clone_args(self, pathfn, recursive):
        args = File._clone_args(self, pathfn, recursive)
        if self.files is None:
            args['files'] = None
        elif recursive:
            args['files'] = [i.clone(pathfn, recursive) for i in self.files]
        else:
            args['files'] = [i for i in self.files]
        return args


class CodeFile(File):
    def __init__(self, path, lang):
        File.__init__(self, path)
        self.lang = lang


class ResourceFile(CodeFile):
    pass


class SourceCodeFile(CodeFile):
    pass


class SourceFile(SourceCodeFile):
    pass


class HeaderFile(SourceCodeFile):
    install_kind = 'data'
    install_root = _InstallRoot.includedir


class PrecompiledHeader(HeaderFile):
    install_kind = None


@_clone_traits(subfiles={'object_file': 'object_path'})
class MsvcPrecompiledHeader(PrecompiledHeader):
    def __init__(self, path, object_path, header_name, format, lang):
        PrecompiledHeader.__init__(self, path, lang)
        self.object_file = ObjectFile(object_path, format, self.lang)
        self.object_file.private = True
        self.header_name = header_name

    def _clone_args(self, pathfn, recursive):
        args = PrecompiledHeader._clone_args(self, pathfn, recursive)
        args['format'] = self.object_file.format
        return args


class HeaderDirectory(Directory):
    install_kind = 'data'
    install_root = _InstallRoot.includedir

    def __init__(self, path, files=None, system=False, langs=None):
        Directory.__init__(self, path, files)
        self.system = system
        self.langs = _listify(langs)


class ModuleDefFile(File):
    pass


class Binary(File):
    install_kind = 'data'
    install_root = _InstallRoot.libdir

    def __init__(self, path, format, lang=None):
        File.__init__(self, path)
        self.format = format
        self.lang = lang


class ObjectFile(Binary):
    pass


# This is used by JVM languages to hold a list of all the object files
# generated by a particular source file's compilation.
class ObjectFileList(ObjectFile):
    install_kind = None

    def __init__(self, path, object_name, format, lang=None):
        ObjectFile.__init__(self, path, format, lang)
        self.object_file = ObjectFile(object_name, format, lang)


# This represents any kind of binary data that's been "linked" (or had some
# similar process applied to it) so that it can be used by a linker/loader,
# installed to the system, etc.
@_clone_traits(exclude={'runtime_deps', 'linktime_deps', 'package_deps'})
class LinkedBinary(Binary):
    def __init__(self, *args, **kwargs):
        Binary.__init__(self, *args, **kwargs)
        self.runtime_deps = []
        self.linktime_deps = []
        self.package_deps = []

    @property
    def install_deps(self):
        return self.runtime_deps + self.linktime_deps


class Executable(LinkedBinary):
    install_kind = 'program'
    install_root = _InstallRoot.bindir


@_clone_traits(exclude={'parent'})
class Library(LinkedBinary):
    @property
    def runtime_file(self):
        return None


# This is used for JVM binaries, which can be both executables and libraries.
# Multiple inheritance is a sign that we should perhaps switch to a trait-based
# system though...
class ExecutableLibrary(Executable, Library):
    install_kind = 'program'
    install_root = _InstallRoot.libdir


class SharedLibrary(Library):
    install_kind = 'program'

    @property
    def runtime_file(self):
        return self


@_clone_traits(exclude={'format', 'lang'})
class LinkLibrary(SharedLibrary):
    def __init__(self, path, library):
        SharedLibrary.__init__(self, path, library.format, library.lang)
        self.library = library
        self.linktime_deps = [library]

    @property
    def runtime_file(self):
        return self.library

    def clone(self, path, recursive=False, inner=None):
        if recursive:
            return self.library.clone(path, True, inner or self)
        return SharedLibrary.clone(self, path, False, inner)


@_clone_traits(subfiles={'soname': 'soname', 'link': 'linkname'})
class VersionedSharedLibrary(SharedLibrary):
    def __init__(self, path, format, lang, soname, linkname):
        SharedLibrary.__init__(self, path, format, lang)
        self.soname = LinkLibrary(soname, self)
        self.link = LinkLibrary(linkname, self.soname)


class StaticLibrary(Library):
    def __init__(self, path, format, lang=None, forward_opts=None):
        Library.__init__(self, path, format, lang)
        self.forward_opts = forward_opts or {}


class WholeArchive(StaticLibrary):
    def __init__(self, library):
        self.library = library

    def __getattribute__(self, name):
        if name in ['library', '_safe_str', '__repr__', '__hash__', '__eq__']:
            return object.__getattribute__(self, name)
        return getattr(object.__getattribute__(self, 'library'), name)


class ExportFile(File):
    private = True


# This refers specifically to DLL files that have an import library, not just
# anything with a .dll extension (for instance, .NET DLLs are just regular
# shared libraries). While this is a "library" in some senses, since you can't
# link to it during building, we just consider it a LinkedBinary.
@_clone_traits(subfiles={'import_lib': 'import_name',
                         'export_file': 'export_name'})
class DllBinary(LinkedBinary):
    install_root = _InstallRoot.bindir
    private = True

    def __init__(self, path, format, lang, import_name, export_name=None):
        LinkedBinary.__init__(self, path, format, lang)
        self.import_lib = LinkLibrary(import_name, self)
        self.export_file = ExportFile(export_name) if export_name else None


class DualUseLibrary(object):
    def __init__(self, shared, static):
        self.shared = shared
        self.static = static
        self.shared.parent = self
        self.static.parent = self

    @property
    def all(self):
        return [self.shared, self.static]

    def __repr__(self):
        return '<DualUseLibrary {!r}>'.format(self.shared.path)

    def __hash__(self):
        return hash(self.shared.path)

    def __eq__(self, rhs):
        return (type(self) == type(rhs) and self.shared == rhs.shared and
                self.static == rhs.static)

    def __ne__(self, rhs):
        return not (self == rhs)

    @property
    def package_deps(self):
        return self.shared.package_deps

    @property
    def install_deps(self):
        return self.shared.install_deps

    @property
    def forward_opts(self):
        return self.static.forward_opts

    def clone(self, *args, **kwargs):
        return DualUseLibrary(self.shared.clone(*args, **kwargs),
                              self.static.clone(*args, **kwargs))


class PkgConfigPcFile(File):
    install_root = _InstallRoot.libdir



# A reference to a macOS framework. Can be used in place of Library objects
# within a Package.


class Framework(object):
    def __init__(self, name, suffix=None):
        self.name = name
        self.suffix = suffix

    @property
    def full_name(self):
        return self.name + ',' + self.suffix if self.suffix else self.name

    def __eq__(self, rhs):
        return (type(self) == type(rhs) and self.name == rhs.name and
                self.suffix == rhs.suffix)

    def __ne__(self, rhs):
        return not (self == rhs)



from . import builtin
from ..backends.make import writer as make
from ..backends.ninja import writer as ninja
from ..build_inputs import build_input
from ..iterutils import unlistify


@build_input('defaults')
class DefaultOutputs(object):
    def __init__(self, build_inputs, env):
        self.default_outputs = []
        self.fallback_defaults = []

    def add(self, output, explicit=False):
        outputs = self.default_outputs if explicit else self.fallback_defaults
        for i in output.all:
            if i.creator:
                outputs.append(i)

    def remove(self, output, explicit=False):
        outputs = self.default_outputs if explicit else self.fallback_defaults
        for i, v in enumerate(outputs):
            if output is v:
                outputs.pop(i)

    @property
    def outputs(self):
        return self.default_outputs or self.fallback_defaults


@builtin.function('build_inputs')
def default(build, *args):
    for i in args:
        build['defaults'].add(i, explicit=True)
    return unlistify(args)


@make.pre_rule
def make_all_rule(build_inputs, buildfile, env):
    buildfile.rule(
        target='all',
        deps=build_inputs['defaults'].outputs,
        phony=True
    )


@ninja.pre_rule
def ninja_all_rule(build_inputs, buildfile, env):
    buildfile.default(['all'])
    buildfile.build(
        output='all',
        rule='phony',
        inputs=build_inputs['defaults'].outputs,
    )



from . import tool
from .common import SimpleCommand
from ..safe_str import shell_literal
from ..shell import shell_list


@tool('bfg9000')
class Bfg9000(SimpleCommand):
    def __init__(self, env):
        SimpleCommand.__init__(self, env, name='bfg9000', env_var='BFG9000',
                               default=env.bfgdir.append('bfg9000'))

    def _call(self, cmd, builddir):
        return cmd + ['refresh', builddir]


@tool('depfixer')
class Depfixer(SimpleCommand):
    def __init__(self, env):
        SimpleCommand.__init__(self, env, name='depfixer', env_var='DEPFIXER',
                               default=env.bfgdir.append('bfg9000-depfixer'))

    def _call(self, cmd, depfile):
        return shell_list(cmd + [shell_literal('<'), depfile,
                                 shell_literal('>>'), depfile])


@tool('jvmoutput')
class JvmOutput(SimpleCommand):
    def __init__(self, env):
        SimpleCommand.__init__(
            self, env, name='jvmoutput', env_var='JVMOUTPUT',
            default=env.bfgdir.append('bfg9000-jvmoutput')
        )

    def _call(self, cmd, output, subcmd):
        return cmd + ['-o', output] + subcmd


@tool('rccdep')
class RccDep(SimpleCommand):
    def __init__(self, env):
        SimpleCommand.__init__(self, env, name='rccdep', env_var='RCCDEP',
                               default=env.bfgdir.append('bfg9000-rccdep'))

    def _call(self, cmd, subcmd, depfile):
        return cmd + subcmd + ['-d', depfile]



import os.path

from . import *


@skip_if_backend('msbuild')
class TestTests(IntegrationTest):
    def __init__(self, *args, **kwargs):
        IntegrationTest.__init__(
            self, os.path.join(examples_dir, '07_tests'), *args, **kwargs
        )

    def test_test(self):
        self.build('test')



import warnings
from itertools import chain, repeat
from six.moves import filter as ifilter

from . import builtin
from .. import shell
from ..backends.make import writer as make
from ..backends.ninja import writer as ninja
from ..build_inputs import Edge
from ..file_types import File, Node, Phony
from ..iterutils import isiterable, iterate, listify
from ..objutils import convert_each
from ..path import Path, Root
from ..safe_str import jbos, safe_str, safe_string
from ..shell import posix as pshell


class Placeholder(safe_string):
    def __init__(self, field, key=slice(None)):
        self._field = field
        self._key = key

    def __getitem__(self, key):
        if self._key != slice(None):
            raise TypeError('{!r} object has already been indexed'
                            .format(type(self).__name__))
        return Placeholder(self._field, key)

    def expand(self, rule):
        data = getattr(rule, self._field)
        if isinstance(self._key, slice):
            return data[self._key]
        else:
            return [data[self._key]]

    @classmethod
    def expand_word(cls, word, rule):
        if isinstance(word, cls):
            return word.expand(rule)
        elif isinstance(word, jbos):
            placeholders = list(ifilter(lambda i: isinstance(i[1], cls),
                                        enumerate(word.bits)))
            if len(placeholders) > 1:
                raise ValueError('only one placeholder per word permitted')
            elif len(placeholders) == 1:
                index, bit = placeholders[0]
                expanded = bit.expand(rule)
                pre, post = word.bits[:index], word.bits[index + 1:]
                return [jbos(*(pre + (safe_str(i),) + post)) for i in expanded]
        return [word]


Input = Placeholder('files')
Output = Placeholder('output')


class BaseCommand(Edge):
    def __init__(self, build, env, name, outputs, cmds, files,
                 environment=None, phony=False, extra_deps=None,
                 description=None):
        self.name = name
        self.files = files
        self.phony = phony

        implicit = [i for line in cmds for i in iterate(line)
                    if isinstance(i, Node) and (i.creator or not phony)]
        implicit.extend(iterate(extra_deps))

        Edge.__init__(self, build, outputs, extra_deps=implicit,
                      description=description)

        # Do this after Edge.__init__ so that self.output is set for our
        # placeholders.
        self.cmds = [env.run_arguments(self._expand_cmd(line))
                     for line in cmds]
        self.env = environment or {}

    @staticmethod
    def convert_args(builtins, kwargs):
        cmd = kwargs.pop('cmd', None)
        cmds = kwargs.pop('cmds', None)
        if (cmd is None) == (cmds is None):
            raise ValueError('exactly one of "cmd" or "cmds" must be ' +
                             'specified')
        kwargs['cmds'] = [cmd] if cmds is None else cmds

        convert_each(kwargs, 'files', builtins['auto_file'])
        return kwargs

    def _expand_cmd(self, cmd):
        if not isiterable(cmd):
            expanded = Placeholder.expand_word(cmd, self)
            if len(expanded) > 1:
                raise ValueError('placeholder can only expand to one item ' +
                                 'when used in a string command line')
            return expanded[0]
        return list(chain.from_iterable(
            Placeholder.expand_word(i, self) for i in cmd)
        )


class Command(BaseCommand):
    console = True

    def __init__(self, build, env, name, **kwargs):
        BaseCommand.__init__(self, build, env, name, Phony(name), phony=True,
                             **kwargs)


@builtin.function('build_inputs', 'builtins', 'env')
def command(build, builtins, env, name, **kwargs):
    kwargs = Command.convert_args(builtins, kwargs)
    return Command(build, env, name, **kwargs).public_output


command.input = Input


class BuildStep(BaseCommand):
    console = False
    msbuild_output = True

    def __init__(self, build, env, name, type=None, args=None, kwargs=None,
                 always_outdated=False, **fwd_kwargs):
        name = listify(name)
        project_name = name[0]

        if not isiterable(type):
            type = repeat(type, len(name))

        if args is None:
            args = repeat([], len(name))
        else:  # pragma: no cover
            # TODO: remove this after 0.5 is released.
            warnings.warn('"args" is deprecated; use "type" instead ' +
                          '(e.g. with a lambda)')

        if kwargs is None:
            kwargs = repeat({}, len(name))
        else:  # pragma: no cover
            # TODO: remove this after 0.5 is released.
            warnings.warn('"kwargs" is deprecated; use "type" instead ' +
                          '(e.g. with a lambda)')

        outputs = [self._make_outputs(*i) for i in
                   zip(name, type, args, kwargs)]

        desc = fwd_kwargs.pop('description', 'build => ' + ' '.join(name))
        BaseCommand.__init__(self, build, env, project_name, outputs,
                             phony=always_outdated, description=desc,
                             **fwd_kwargs)

    @staticmethod
    def convert_args(builtins, kwargs):
        if kwargs.get('type') is None:
            kwargs['type'] = builtins['auto_file']
        return BaseCommand.convert_args(builtins, kwargs)

    @staticmethod
    def _make_outputs(name, type, args, kwargs):
        result = type(Path(name, Root.builddir), *args, **kwargs)
        if not isinstance(result, File):
            raise ValueError('expected a function returning a file')
        return result


@builtin.function('build_inputs', 'builtins', 'env')
def build_step(build, builtins, env, name, **kwargs):
    kwargs = BuildStep.convert_args(builtins, kwargs)
    return BuildStep(build, env, name, **kwargs).public_output


build_step.input = Input
build_step.output = Output


@make.rule_handler(Command, BuildStep)
def make_command(rule, build_inputs, buildfile, env):
    # Join all the commands onto one line so that users can use 'cd' and such.
    make.multitarget_rule(
        buildfile,
        targets=rule.output,
        deps=rule.files + rule.extra_deps,
        order_only=(make.directory_deps(rule.output) if
                    isinstance(rule, BuildStep) else []),
        recipe=[pshell.global_env(rule.env, rule.cmds)],
        phony=rule.phony
    )


@ninja.rule_handler(Command, BuildStep)
def ninja_command(rule, build_inputs, buildfile, env):
    ninja.command_build(
        buildfile, env,
        output=rule.output,
        inputs=rule.files,
        implicit=rule.extra_deps,
        command=shell.global_env(rule.env, rule.cmds),
        console=rule.console,
        phony=rule.phony,
        description=rule.description
    )


try:
    from ..backends.msbuild import writer as msbuild

    @msbuild.rule_handler(Command, BuildStep)
    def msbuild_command(rule, build_inputs, solution, env):
        command = msbuild.CommandProject.convert_command(shell.global_env(
            rule.env, rule.cmds
        ))
        project = msbuild.CommandProject(
            env, name=rule.name,
            commands=[msbuild.CommandProject.task(
                'Exec', Command=command, WorkingDirectory='$(OutDir)'
            )],
            dependencies=solution.dependencies(rule.extra_deps),
        )
        solution[rule.output[0]] = project
except ImportError:  # pragma: no cover
    pass



from enum import Enum
from six import string_types
from six.moves import cStringIO as StringIO

from .. import iterutils
from .. import path
from .. import safe_str
from ..shell import posix as pshell

Syntax = Enum('Syntax', ['variable', 'shell'])


class Writer(object):
    def __init__(self, stream):
        self.stream = stream

    def write_literal(self, string):
        self.stream.write(string)

    def write(self, thing, syntax, shell_quote=pshell.quote_info):
        thing = safe_str.safe_str(thing)
        shelly = syntax == Syntax.shell
        escaped = False

        if isinstance(thing, safe_str.literal_types):
            escaped = True
            self.write_literal(thing.string)
        elif isinstance(thing, string_types):
            if shelly and shell_quote:
                thing, escaped = shell_quote(thing)
            self.write_literal(thing)
        elif isinstance(thing, safe_str.jbos):
            for i in thing.bits:
                escaped |= self.write(i, syntax, shell_quote)
        elif isinstance(thing, path.BasePath):
            out = Writer(StringIO())
            thing = thing.realize(path_vars, shelly)
            escaped = out.write(thing, syntax, pshell.inner_quote_info)

            thing = out.stream.getvalue()
            if shelly and escaped:
                thing = pshell.wrap_quotes(thing)
            self.write_literal(thing)
        else:
            raise TypeError(type(thing))

        return escaped

    def write_each(self, things, syntax, delim=safe_str.literal(' '),
                   prefix=None, suffix=None, shell_quote=pshell.quote_info):
        for i in iterutils.tween(things, delim, prefix, suffix):
            self.write(i, syntax, shell_quote)


class Variable(safe_str.safe_string_ops):
    def __init__(self, name):
        self.name = name

    def use(self):
        return safe_str.literal('${{{}}}'.format(self.name))

    def _safe_str(self):
        return self.use()

    def __str__(self):
        raise NotImplementedError()

    def __repr__(self):
        return repr(self.use())

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, rhs):
        return self.name == rhs.name

    def __ne__(self, rhs):
        return not (self == rhs)


path_vars = {i: Variable(i.name) for i in path.InstallRoot}



from .posix import PosixHostPlatform, PosixTargetPlatform


class CygwinHostPlatform(PosixHostPlatform):
    pass


class CygwinTargetPlatform(PosixTargetPlatform):
    @property
    def object_format(self):
        return 'coff'

    @property
    def executable_ext(self):
        return '.exe'

    @property
    def shared_library_ext(self):
        return '.dll'

    @property
    def has_import_library(self):
        return True

    @property
    def has_versioned_library(self):
        return False



from itertools import chain

from . import builder
from .. import options as opts, safe_str, shell
from .common import BuildCommand, Builder, choose_builder
from ..file_types import HeaderFile, SourceFile
from ..iterutils import iterate
from ..languages import known_langs
from ..path import Path
from ..versioning import detect_version

with known_langs.make('qtmoc', base='c++') as x:
    x.vars(compiler='MOC', flags='MOCFLAGS')

with known_langs.make('qrc') as x:
    x.vars(compiler='RCC', flags='RCCFLAGS')
    x.exts(resource=['.qrc'])

with known_langs.make('qtui') as x:
    x.vars(compiler='UIC', flags='UICFLAGS')
    x.exts(resource=['.ui'])


@builder('qtmoc')
def moc_builder(env):
    return choose_builder(env, known_langs['qtmoc'], ['moc'], (MocBuilder,))


class MocBuilder(Builder):
    def __init__(self, env, langinfo, command, version_output):
        Builder.__init__(self, langinfo.name,
                         *self._parse_brand(version_output))

        name = langinfo.var('compiler').lower()
        mocflags_name = langinfo.var('flags').lower()
        mocflags = shell.split(env.getvar(langinfo.var('flags'), ''))

        self.transpiler = MocCompiler(self, env, name, name, command,
                                      flags=(mocflags_name, mocflags))

    @staticmethod
    def _parse_brand(version_output):
        if 'moc' in version_output:
            return 'qt', detect_version(version_output)
        return 'unknown', None

    @staticmethod
    def check_command(env, command):
        return env.execute(command + ['--version'], stdout=shell.Mode.pipe,
                           stderr=shell.Mode.devnull)


class MocCompiler(BuildCommand):
    @property
    def deps_flavor(self):
        return None

    @property
    def needs_libs(self):
        return False

    def _call(self, cmd, input, output, flags=None):
        return list(chain(
            cmd, iterate(flags), [input, '-o', output]
        ))

    def default_name(self, input, context):
        if isinstance(input, SourceFile):
            return input.path.stripext('.moc').suffix
        base, leaf = input.path.stripext(
            known_langs['c++'].default_ext('source')
        ).splitleaf()
        return base.append('moc_' + leaf).suffix

    def output_file(self, name, context):
        return SourceFile(Path(name), 'c++')

    def flags(self, options, output=None, mode='normal'):
        flags = []
        for i in options:
            if isinstance(i, opts.include_dir):
                flags.append('-I' + i.directory.path)
            elif isinstance(i, opts.define):
                if i.value:
                    flags.append('-D' + i.name + '=' + i.value)
                else:
                    flags.append('-D' + i.name)
            elif isinstance(i, opts.warning):
                for j in i.value:
                    if j == opts.WarningValue.disable:
                        flags.append('--no-warnings')
                    else:
                        raise ValueError('unsupported warning level {!r}'
                                         .format(j))
            elif isinstance(i, safe_str.stringy_types):
                flags.append(i)
            else:
                raise TypeError('unknown option type {!r}'.format(type(i)))
        return flags


@builder('qrc')
def qrc_builder(env):
    return choose_builder(env, known_langs['qrc'], ['rcc'], (RccBuilder,))


class RccBuilder(Builder):
    def __init__(self, env, langinfo, command, version_output):
        Builder.__init__(self, langinfo.name,
                         *self._parse_brand(version_output))

        name = langinfo.var('compiler').lower()
        rccflags_name = langinfo.var('flags').lower()
        rccflags = shell.split(env.getvar(langinfo.var('flags'), ''))

        self.transpiler = RccCompiler(self, env, name, command, rccflags_name,
                                      rccflags)

    @staticmethod
    def _parse_brand(version_output):
        if 'rcc' in version_output:
            return 'qt', detect_version(version_output)
        return 'unknown', None

    @staticmethod
    def check_command(env, command):
        return env.execute(command + ['--version'], stdout=shell.Mode.pipe,
                           stderr=shell.Mode.devnull)


class RccCompiler(BuildCommand):
    def __init__(self, builder, env, name, command, rccflags_name, rccflags):
        BuildCommand.__init__(self, builder, env, name, name, command,
                              flags=(rccflags_name, rccflags))

    @property
    def deps_flavor(self):
        return 'gcc'

    @property
    def needs_libs(self):
        return False

    def _call(self, cmd, input, output, deps=None, flags=None):
        result = list(chain(cmd, iterate(flags), [input, '-o', output]))
        if deps:
            return self.env.tool('rccdep')(result, deps)
        return result

    def default_name(self, input, context):
        return input.path.stripext(
            known_langs['c++'].default_ext('source')
        ).suffix

    def output_file(self, name, context):
        return SourceFile(Path(name), 'c++')

    def flags(self, options, output=None, mode='normal'):
        flags = []
        for i in options:
            if isinstance(i, safe_str.stringy_types):
                flags.append(i)
            else:
                raise TypeError('unknown option type {!r}'.format(type(i)))
        return flags


@builder('qtui')
def qtui_builder(env):
    return choose_builder(env, known_langs['qtui'], ['uic'], (UicBuilder,))


class UicBuilder(Builder):
    def __init__(self, env, langinfo, command, version_output):
        Builder.__init__(self, langinfo.name,
                         *self._parse_brand(version_output))

        name = langinfo.var('compiler').lower()
        uicflags_name = langinfo.var('flags').lower()
        uicflags = shell.split(env.getvar(langinfo.var('flags'), ''))

        self.transpiler = UicCompiler(self, env, name, command, uicflags_name,
                                      uicflags)

    @staticmethod
    def _parse_brand(version_output):
        if 'uic' in version_output:
            return 'qt', detect_version(version_output)
        return 'unknown', None

    @staticmethod
    def check_command(env, command):
        return env.execute(command + ['--version'], stdout=shell.Mode.pipe,
                           stderr=shell.Mode.devnull)


class UicCompiler(BuildCommand):
    def __init__(self, builder, env, name, command, uicflags_name, uicflags):
        BuildCommand.__init__(self, builder, env, name, name, command,
                              flags=(uicflags_name, uicflags))

    @property
    def deps_flavor(self):
        return None

    @property
    def needs_libs(self):
        return False

    def _call(self, cmd, input, output, flags=None):
        return list(chain(
            cmd, iterate(flags), [input, '-o', output]
        ))

    def default_name(self, input, context):
        base, leaf = input.path.stripext('.h').splitleaf()
        return base.append('ui_' + leaf).suffix

    def output_file(self, name, context):
        return HeaderFile(Path(name), 'c++')

    def flags(self, options, output=None, mode='normal'):
        flags = []
        for i in options:
            if isinstance(i, safe_str.stringy_types):
                flags.append(i)
            else:
                raise TypeError('unknown option type {!r}'.format(type(i)))
        return flags



import colorama
import logging
import os
import sys
import traceback
import warnings
from logging import getLogger, CRITICAL, ERROR, WARNING, INFO, DEBUG  # noqa


class UserDeprecationWarning(DeprecationWarning):
    pass


def _is_bfg_src(filename):
    try:
        rel = os.path.relpath(filename, os.path.dirname(__file__))
    except ValueError:
        return False
    return not rel.startswith(os.pardir + os.sep)


def _filter_stack(stack):
    # Find where the user's stack frames begin and end.
    gen = enumerate(stack)
    for start, line in gen:
        if not _is_bfg_src(line[0]):
            break
    else:
        start = len(stack)

    for end, line in gen:
        if _is_bfg_src(line[0]):
            break
    else:
        end = len(stack)

    return stack[:start], stack[start:end], stack[end:]


def _format_stack(stack):
    if len(stack) == 0:
        return ''
    # Put the newline at the beginning, since this helps our formatting later.
    return '\n' + ''.join(traceback.format_list(stack)).rstrip()


class StackFilter(object):
    def __init__(self, has_stack=True):
        self.has_stack = has_stack

    def filter(self, record):
        has_stack = bool((record.exc_info and record.exc_info[0]) or
                         getattr(record, 'show_stack', False))
        return has_stack == self.has_stack


class StackfulStreamHandler(logging.StreamHandler):
    def __init__(self, *args, **kwargs):
        self.debug = kwargs.pop('debug', False)
        logging.StreamHandler.__init__(self, *args, **kwargs)

    def emit(self, record):
        if record.exc_info:
            if isinstance(record.exc_info[1], SyntaxError):
                e = record.exc_info[1]
                record.msg = e.msg

                # Figure out where to put the carat.
                text = e.text.expandtabs()
                dedent = len(text) - len(text.lstrip())
                offset = 4 - dedent - 1 + e.offset

                record.full_stack = [(e.filename, e.lineno, '<module>',
                                      e.text + ' ' * offset + '^')]
            else:
                record.full_stack = traceback.extract_tb(record.exc_info[2])
            record.exc_info = None

        pre, stack, post = _filter_stack(record.full_stack)

        record.stack_pre = _format_stack(pre)
        record.stack = _format_stack(stack)
        record.stack_post = _format_stack(post)

        if len(stack):
            record.user_pathname = stack[-1][0]
            record.user_lineno = stack[-1][1]
        else:
            record.user_pathname = record.pathname
            record.user_lineno = record.lineno

        if len(stack) or self.debug:
            return logging.StreamHandler.emit(self, record)
        logging.root.handle(record)


def init(color='auto', debug=False, warn_once=False, stream=None):
    if color == 'always':
        colorama.init(strip=False)
    elif color == 'never':
        colorama.init(strip=True, convert=False)
    else:  # color == 'auto'
        colorama.init()

    warnings.filterwarnings('default', category=UserDeprecationWarning)
    if warn_once:
        warnings.filterwarnings('once')

    logging.addLevelName(logging.CRITICAL,
                         '\033[1;41;37m' + 'critical' + '\033[0m')
    logging.addLevelName(logging.ERROR, '\033[1;31m' + 'error' + '\033[0m')
    logging.addLevelName(logging.WARNING, '\033[1;33m' + 'warning' + '\033[0m')
    logging.addLevelName(logging.INFO, '\033[1;34m' + 'info' + '\033[0m')
    logging.addLevelName(logging.DEBUG, '\033[1;35m' + 'debug' + '\033[0m')

    logging.root.setLevel(logging.DEBUG if debug else logging.INFO)

    stackless = logging.StreamHandler(stream)
    stackless.addFilter(StackFilter(has_stack=False))

    fmt = '%(levelname)s: %(message)s'

    stackless.setFormatter(logging.Formatter(fmt))
    logging.root.addHandler(stackless)

    stackful = StackfulStreamHandler(stream, debug=debug)
    stackful.addFilter(StackFilter(has_stack=True))

    fmt = '%(levelname)s: %(user_pathname)s:%(user_lineno)d: %(message)s'
    if debug:
        fmt += '\033[2m%(stack_pre)s\033[0m'
    fmt += '%(stack)s'
    if debug:
        fmt += '\033[2m%(stack_post)s\033[0m'

    stackful.setFormatter(logging.Formatter(fmt))
    logging.root.addHandler(stackful)


def log_stack(level, message, *args, **kwargs):
    stacklevel = kwargs.pop('stacklevel', 0)
    extra = {
        'full_stack': traceback.extract_stack()[1:-1 - stacklevel],
        'show_stack': kwargs.pop('show_stack', True)
    }
    logging.log(level, message, *args, extra=extra, **kwargs)


def info(msg, show_stack=False):
    log_stack(INFO, msg, show_stack=show_stack, stacklevel=1)


def debug(msg, show_stack=True):
    log_stack(DEBUG, msg, show_stack=show_stack, stacklevel=1)


def _showwarning(message, category, filename, lineno, file=None, line=None):
    # Python 3.6 changes how stacklevel is counted.
    stacklevel = 2 if sys.version_info >= (3, 6) else 1
    log_stack(logging.WARNING, message, stacklevel=stacklevel)


warnings.showwarning = _showwarning



import os.path
import re
from itertools import chain

from . import pkg_config
from .common import BuildCommand, Builder, check_which, library_macro
from .. import options as opts, safe_str, shell
from ..arguments.windows import ArgumentParser
from ..builtins.file_types import make_immediate_file
from ..exceptions import PackageResolutionError
from ..file_types import *
from ..iterutils import (default_sentinel, iterate, listify, merge_into_dict,
                         uniques)
from ..languages import known_langs, known_formats
from ..objutils import memoize
from ..packages import CommonPackage, Framework, PackageKind
from ..path import abspath, exists, Path, Root
from ..versioning import detect_version, SpecifierSet

_warning_flags = {
    opts.WarningValue.disable: '/w',
    opts.WarningValue.all    : '/W3',
    opts.WarningValue.extra  : '/W4',
    opts.WarningValue.error  : '/WX',
}

_optimize_flags = {
    opts.OptimizeValue.disable : '/Od',
    opts.OptimizeValue.size    : '/O1',
    opts.OptimizeValue.speed   : '/O2',
    opts.OptimizeValue.linktime: '/GL',
}


class MsvcBuilder(Builder):
    def __init__(self, env, langinfo, command, version_output):
        Builder.__init__(self, langinfo.name,
                         *self._parse_brand(version_output))
        self.object_format = env.target_platform.object_format

        name = langinfo.var('compiler').lower()
        ldinfo = known_formats['native', 'dynamic']
        arinfo = known_formats['native', 'static']

        # Look for the last argument that looks like our compiler and use its
        # directory as the base directory to find the linkers.
        origin = ''
        for i in reversed(command):
            if os.path.basename(i) in ('cl', 'cl.exe'):
                origin = os.path.dirname(i)
        link_command = check_which(
            env.getvar(ldinfo.var('linker'), os.path.join(origin, 'link')),
            env.variables, kind='dynamic linker'.format(self.lang)
        )
        lib_command = check_which(
            env.getvar(arinfo.var('linker'), os.path.join(origin, 'lib')),
            env.variables, kind='static linker'.format(self.lang)
        )

        cflags_name = langinfo.var('flags').lower()
        cflags = (
            shell.split(env.getvar('CPPFLAGS', '')) +
            shell.split(env.getvar(langinfo.var('flags'), ''))
        )

        ld_name = ldinfo.var('linker').lower()
        ldflags_name = ldinfo.var('flags').lower()
        ldflags = shell.split(env.getvar(ldinfo.var('flags'), ''))
        ldlibs_name = ldinfo.var('libs').lower()
        ldlibs = shell.split(env.getvar(ldinfo.var('libs'), ''))

        ar_name = arinfo.var('linker').lower()
        arflags_name = arinfo.var('flags').lower()
        arflags = shell.split(env.getvar(arinfo.var('flags'), ''))

        self.compiler = MsvcCompiler(self, env, name, command, cflags_name,
                                     cflags)
        self.pch_compiler = MsvcPchCompiler(self, env, name, command,
                                            cflags_name, cflags)
        self._linkers = {
            'executable': MsvcExecutableLinker(
                self, env, name, ld_name, link_command, ldflags_name, ldflags,
                ldlibs_name, ldlibs
            ),
            'shared_library': MsvcSharedLibraryLinker(
                self, env, name, ld_name, link_command, ldflags_name, ldflags,
                ldlibs_name, ldlibs
            ),
            'static_library': MsvcStaticLinker(
                self, env, ar_name, lib_command, arflags_name, arflags
            ),
        }
        self.packages = MsvcPackageResolver(self, env)
        self.runner = None

    @staticmethod
    def _parse_brand(version_output):
        if 'Microsoft (R)' in version_output:
            return 'msvc', detect_version(version_output)
        # XXX: Detect clang-cl.
        return 'unknown', None

    @staticmethod
    def check_command(env, command):
        return env.execute(command + ['/?'], stdout=shell.Mode.pipe,
                           stderr=shell.Mode.stdout)

    @property
    def flavor(self):
        return 'msvc'

    @property
    def family(self):
        return 'native'

    @property
    def auto_link(self):
        return True

    @property
    def can_dual_link(self):
        return False

    def linker(self, mode):
        return self._linkers[mode]


class MsvcBaseCompiler(BuildCommand):
    def __init__(self, builder, env, rule_name, command_var, command,
                 cflags_name, cflags):
        BuildCommand.__init__(self, builder, env, rule_name, command_var,
                              command, flags=(cflags_name, cflags))

    @property
    def deps_flavor(self):
        return 'msvc'

    @property
    def needs_libs(self):
        return False

    def search_dirs(self, strict=False):
        cpath = [abspath(i) for i in
                 self.env.getvar('CPATH', '').split(os.pathsep)]
        include = [abspath(i) for i in
                   self.env.getvar('INCLUDE', '').split(os.pathsep)]
        return cpath + include

    def _call(self, cmd, input, output, deps=None, flags=None):
        result = list(chain( cmd, self._always_flags, iterate(flags) ))
        if deps:
            result.append('/showIncludes')
        result.extend(['/c', input, '/Fo' + output])
        return result

    @property
    def _always_flags(self):
        return ['/nologo', '/EHsc']

    def flags(self, options, output=None, mode='normal'):
        syntax = 'cc' if mode == 'pkg-config' else 'msvc'
        flags = []
        for i in options:
            if isinstance(i, opts.include_dir):
                prefix = '-I' if syntax == 'cc' else '/I'
                flags.append(prefix + i.directory.path)
            elif isinstance(i, opts.define):
                prefix = '-D' if syntax == 'cc' else '/D'
                if i.value:
                    flags.append(prefix + i.name + '=' + i.value)
                else:
                    flags.append(prefix + i.name)
            elif isinstance(i, opts.std):
                flags.append('/std:' + i.value)
            elif isinstance(i, opts.warning):
                for j in i.value:
                    flags.append(_warning_flags[j])
            elif isinstance(i, opts.debug):
                flags.append('/Zi')
            elif isinstance(i, opts.optimize):
                for j in i.value:
                    flags.append(_optimize_flags[j])
            elif isinstance(i, opts.pch):
                flags.append('/Yu' + i.header.header_name)
            elif isinstance(i, opts.sanitize):
                flags.append('/RTC1')
            elif isinstance(i, safe_str.stringy_types):
                flags.append(i)
            else:
                raise TypeError('unknown option type {!r}'.format(type(i)))
        return flags

    @staticmethod
    @memoize
    def __parser():
        parser = ArgumentParser()
        parser.add('/nologo')
        parser.add('/D', '-D', type=list, dest='defines')
        parser.add('/I', '-I', type=list, dest='includes')

        warn = parser.add('/W', type=dict, dest='warnings')
        warn.add('0', '1', '2', '3', '4', 'all', dest='level')
        warn.add('X', type=bool, dest='as_error')
        warn.add('X-', type=bool, dest='as_error', value=False)
        parser.add('/w', type='alias', base=warn, value='0')

        pch = parser.add('/Y', type=dict, dest='pch')
        pch.add('u', type=str, dest='use')
        pch.add('c', type=str, dest='create')

        return parser

    def parse_flags(self, flags):
        result, extra = self.__parser().parse_known(flags)
        result['extra'] = extra
        return result


class MsvcCompiler(MsvcBaseCompiler):
    def __init__(self, builder, env, name, command, cflags_name, cflags):
        MsvcBaseCompiler.__init__(self, builder, env, name, name, command,
                                  cflags_name, cflags)

    @property
    def accepts_pch(self):
        return True

    def default_name(self, input, context):
        return input.path.stripext().suffix

    def output_file(self, name, context):
        pch = getattr(context, 'pch', None)
        output = ObjectFile(Path(name + '.obj'),
                            self.builder.object_format, self.lang)
        if pch:
            output.extra_objects = [pch.object_file]
        return output


class MsvcPchCompiler(MsvcBaseCompiler):
    def __init__(self, builder, env, name, command, cflags_name, cflags):
        MsvcBaseCompiler.__init__(self, builder, env, name + '_pch', name,
                                  command, cflags_name, cflags)

    @property
    def num_outputs(self):
        return 2

    @property
    def accepts_pch(self):
        # You can't to pass a PCH to a PCH compiler!
        return False

    def _call(self, cmd, input, output, deps=None, flags=None):
        output = listify(output)
        result = MsvcBaseCompiler._call(self, cmd, input, output[1], deps,
                                        flags)
        result.append('/Fp' + output[0])
        return result

    def pre_build(self, build, name, context):
        header = getattr(context, 'file')
        options = opts.option_list()

        if context.pch_source is None:
            ext = known_langs[self.lang].default_ext('source')
            context.pch_source = SourceFile(header.path.stripext(ext).reroot(),
                                            header.lang)
            with make_immediate_file(build, self.env, context.pch_source) \
                 as out:  # noqa
                out.write('#include "{}"\n'.format(header.path.basename()))

            # Add the include path for the header to ensure the PCH source
            # finds it.
            d = HeaderDirectory(header.path.parent())
            options.append(opts.include_dir(d))

        # Add flag to create PCH file.
        options.append('/Yc' + header.path.suffix)
        return options

    def default_name(self, input, context):
        return input.path.suffix

    def output_file(self, name, context):
        pchpath = Path(name).stripext('.pch')
        objpath = context.pch_source.path.stripext('.obj').reroot()
        output = MsvcPrecompiledHeader(
            pchpath, objpath, name, self.builder.object_format, self.lang
        )
        return [output, output.object_file]


class MsvcLinker(BuildCommand):
    __lib_re = re.compile(r'(.*)\.lib$')
    __allowed_langs = {
        'c'     : {'c'},
        'c++'   : {'c', 'c++'},
    }

    def __init__(self, builder, env, rule_name, name, command, ldflags_name,
                 ldflags, ldlibs_name, ldlibs):
        BuildCommand.__init__(
            self, builder, env, rule_name, name, command,
            flags=(ldflags_name, ldflags), libs=(ldlibs_name, ldlibs)
        )

    def _extract_lib_name(self, library):
        basename = library.path.basename()
        m = self.__lib_re.match(basename)
        if not m:
            raise ValueError("'{}' is not a valid library name"
                             .format(basename))
        return m.group(1)

    def can_link(self, format, langs):
        return (format == self.builder.object_format and
                self.__allowed_langs[self.lang].issuperset(langs))

    @property
    def needs_libs(self):
        return True

    @property
    def has_link_macros(self):
        return True

    def search_dirs(self, strict=False):
        lib_path = [abspath(i) for i in
                    self.env.getvar('LIBRARY_PATH', '').split(os.pathsep)]
        lib = [abspath(i) for i in
               self.env.getvar('LIB', '').split(os.pathsep)]
        return lib_path + lib

    def _call(self, cmd, input, output, libs=None, flags=None):
        return list(chain(
            cmd, self._always_flags, iterate(flags), iterate(input),
            iterate(libs), ['/OUT:' + output]
        ))

    @property
    def _always_flags(self):
        return ['/nologo']

    def always_libs(self, primary):
        return opts.option_list()

    def _link_lib(self, library, syntax):
        if isinstance(library, Framework):
            raise TypeError('MSVC does not support frameworks')
        elif isinstance(library, WholeArchive):
            if not self.version or self.version in SpecifierSet('>=19'):
                return ['/WHOLEARCHIVE:' + library.path]
            raise TypeError('whole-archives require MSVC 2015 Update 2')

        # Unlike the cc linker, we only support Library objects here (strings
        # aren't allowed!)
        if syntax == 'cc':
            return ['-l' + self._extract_lib_name(library)]
        else:
            # Pass the raw path to the library. We do this to avoid adding more
            # `/LIBPATH` options than we really need, which makes it easier to
            # find the right library when there are name collisions (e.g.
            # linking to a system `libfoo` when also building a local `libfoo`
            # to use elsewhere).
            return [library.path]

    def _lib_dir(self, library, syntax):
        if syntax == 'cc' and not isinstance(library, WholeArchive):
            return [library.path.parent()]
        return []

    def flags(self, options, output=None, mode='normal'):
        syntax = 'cc' if mode == 'pkg-config' else 'msvc'
        flags, lib_dirs = [], []
        for i in options:
            if isinstance(i, opts.lib_dir):
                lib_dirs.append(i.directory.path)
            elif isinstance(i, opts.lib):
                lib_dirs.extend(self._lib_dir(i.library, syntax))
            elif isinstance(i, opts.module_def):
                flags.append('/DEF:' + i.value.path)
            elif isinstance(i, opts.debug):
                flags.append('/DEBUG')
            elif isinstance(i, opts.optimize):
                if opts.OptimizeValue.linktime in i.value:
                    flags.append('/LTCG')
            elif isinstance(i, safe_str.stringy_types):
                flags.append(i)
            elif isinstance(i, opts.lib_literal):
                pass
            else:
                raise TypeError('unknown option type {!r}'.format(type(i)))

        prefix = '-L' if syntax == 'cc' else '/LIBPATH:'
        flags.extend(prefix + i for i in uniques(lib_dirs))
        return flags

    def lib_flags(self, options, mode='normal'):
        syntax = 'cc' if mode == 'pkg-config' else 'msvc'
        flags = []
        for i in options:
            if isinstance(i, opts.lib):
                flags.extend(self._link_lib(i.library, syntax))
            elif isinstance(i, opts.lib_literal):
                flags.append(i.value)
        return flags

    @staticmethod
    @memoize
    def __parser():
        parser = ArgumentParser()
        parser.add('/nologo')
        return parser

    @staticmethod
    @memoize
    def __lib_parser():
        parser = ArgumentParser()
        parser.add('/nologo')
        parser.add_unnamed('libs')
        return parser

    def parse_flags(self, flags, lib_flags):
        result, extra = self.__parser().parse_known(flags)
        libresult, libextra = self.__lib_parser().parse_known(lib_flags)

        merge_into_dict(result, libresult)
        result['extra'] = extra + libextra
        return result


class MsvcExecutableLinker(MsvcLinker):
    def __init__(self, builder, env, name, command_var, command, ldflags_name,
                 ldflags, ldlibs_name, ldlibs):
        MsvcLinker.__init__(self, builder, env, name + '_link', command_var,
                            command, ldflags_name, ldflags, ldlibs_name,
                            ldlibs)

    def output_file(self, name, context):
        path = Path(name + self.env.target_platform.executable_ext)
        return Executable(path, self.builder.object_format, self.lang)


class MsvcSharedLibraryLinker(MsvcLinker):
    def __init__(self, builder, env, name, command_var, command, ldflags_name,
                 ldflags, ldlibs_name, ldlibs):
        MsvcLinker.__init__(self, builder, env, name + '_linklib', command_var,
                            command, ldflags_name, ldflags, ldlibs_name,
                            ldlibs)

    @property
    def num_outputs(self):
        return 2

    def _call(self, cmd, input, output, libs=None, flags=None):
        result = MsvcLinker._call(self, cmd, input, output[0], libs, flags)
        result.append('/IMPLIB:' + output[1])
        return result

    @property
    def _always_flags(self):
        return MsvcLinker._always_flags.fget(self) + ['/DLL']

    def compile_options(self, context):
        options = opts.option_list()
        if self.has_link_macros:
            options.append(opts.define(library_macro(
                context.name, 'shared_library'
            )))
        return options

    def output_file(self, name, context):
        dllname = Path(name + self.env.target_platform.shared_library_ext)
        impname = Path(name + '.lib')
        expname = Path(name + '.exp')
        dll = DllBinary(dllname, self.builder.object_format, self.lang,
                        impname, expname)
        return [dll, dll.import_lib, dll.export_file]


class MsvcStaticLinker(BuildCommand):
    def __init__(self, builder, env, name, command, arflags_name, arflags):
        BuildCommand.__init__(self, builder, env, name, name,
                              command, flags=(arflags_name, arflags))

    def can_link(self, format, langs):
        return format == self.builder.object_format

    @property
    def has_link_macros(self):
        return True

    def _call(self, cmd, input, output, flags=None):
        return list(chain(
            cmd, iterate(flags), iterate(input), ['/OUT:' + output]
        ))

    def compile_options(self, context):
        return self.forwarded_compile_options(context)

    def forwarded_compile_options(self, context):
        options = opts.option_list()
        if self.has_link_macros:
            options.append(opts.define(library_macro(
                context.name, 'static_library'
            )))
        return options

    def flags(self, options, output=None, mode='normal'):
        flags = []
        for i in options:
            if isinstance(i, safe_str.stringy_types):
                flags.append(i)
            else:
                raise TypeError('unknown option type {!r}'.format(type(i)))
        return flags

    def output_file(self, name, context):
        return StaticLibrary(Path(name + '.lib'),
                             self.builder.object_format, context.langs)

    def parse_flags(self, flags):
        return {'extra': flags}


class MsvcPackageResolver(object):
    def __init__(self, builder, env):
        self.builder = builder
        self.env = env

        self.include_dirs = [i for i in uniques(chain(
            self.builder.compiler.search_dirs(),
            self.env.host_platform.include_dirs
        )) if exists(i)]

        self.lib_dirs = [i for i in uniques(chain(
            self.builder.linker('executable').search_dirs(),
            self.env.host_platform.lib_dirs
        )) if exists(i)]

    @property
    def lang(self):
        return self.builder.lang

    def header(self, name, search_dirs=None):
        if search_dirs is None:
            search_dirs = self.include_dirs

        for base in search_dirs:
            if base.root != Root.absolute:
                raise ValueError('expected an absolute path')
            if exists(base.append(name)):
                return HeaderDirectory(base, None, system=True)

        raise PackageResolutionError("unable to find header '{}'".format(name))

    def library(self, name, kind=PackageKind.any, search_dirs=None):
        if search_dirs is None:
            search_dirs = self.lib_dirs
        libname = name + '.lib'

        for base in search_dirs:
            if base.root != Root.absolute:
                raise ValueError('expected an absolute path')
            fullpath = base.append(libname)
            if exists(fullpath):
                # We don't actually know what kind of library this is. It could
                # be a static library or an import library (which we classify
                # as a kind of shared lib).
                return Library(fullpath, self.builder.object_format)
        raise PackageResolutionError("unable to find library '{}'"
                                     .format(name))

    def resolve(self, name, version, kind, headers, lib_names):
        format = self.builder.object_format
        try:
            return pkg_config.resolve(self.env, name, format, version, kind)
        except (OSError, PackageResolutionError):
            if lib_names is default_sentinel:
                lib_names = self.env.target_platform.transform_package(name)

            compile_options = opts.option_list(
                opts.include_dir(self.header(i)) for i in iterate(headers)
            )
            link_options = opts.option_list(
                opts.lib(self.library(i, kind)) for i in iterate(lib_names)
            )
            return CommonPackage(name, format, compile_options, link_options)



import re
from collections import namedtuple, OrderedDict
from enum import Enum
from six import iteritems, string_types
from six.moves import cStringIO as StringIO

from ... import path
from ... import safe_str
from ... import shell
from ... import iterutils
from ...platforms.host import platform_info
from ...tools.common import Command
from ...versioning import Version

__all__ = ['NinjaFile', 'Section', 'Syntax', 'Writer', 'Variable', 'var',
           'path_vars']

Rule = namedtuple('Rule', ['command', 'depfile', 'deps', 'description',
                           'generator', 'pool', 'restat'])
Build = namedtuple('Build', ['outputs', 'rule', 'inputs', 'implicit',
                             'order_only', 'variables'])

Syntax = Enum('Syntax', ['output', 'input', 'shell', 'clean'])
Section = Enum('Section', ['path', 'command', 'flags', 'other'])

_comment_tmpl = """
# Do not edit this file! It was automatically generated by bfg9000.
# Instead, you should edit the source file that created this:
# {}
""".strip()


class Writer(object):
    def __init__(self, stream, shell=shell):
        self.stream = stream
        self.shell = shell

    @staticmethod
    def escape_str(string, syntax):
        if '\n' in string:
            raise ValueError('illegal newline')

        if syntax == Syntax.output:
            return re.sub(r'([:$ ])', r'$\1', string)
        elif syntax == Syntax.input:
            return re.sub(r'([$ ])', r'$\1', string)
        elif syntax in [Syntax.shell, Syntax.clean]:
            return string.replace('$', '$$')

        raise ValueError(
            'unknown syntax {!r}'.format(syntax)
        )  # pragma: no cover

    def write_literal(self, string):
        self.stream.write(string)

    def write(self, thing, syntax, shell_quote=iterutils.default_sentinel):
        if shell_quote is iterutils.default_sentinel:
            shell_quote = self.shell.quote_info
        thing = safe_str.safe_str(thing)
        shelly = syntax == Syntax.shell
        escaped = False

        if isinstance(thing, safe_str.literal):
            escaped = True
            self.write_literal(thing.string)
        elif isinstance(thing, safe_str.shell_literal):
            escaped = True
            self.write_literal(self.escape_str(thing.string, syntax))
        elif isinstance(thing, string_types):
            if shelly and shell_quote:
                thing, escaped = shell_quote(thing)
            self.write_literal(self.escape_str(thing, syntax))
        elif isinstance(thing, safe_str.jbos):
            for i in thing.bits:
                escaped |= self.write(i, syntax, shell_quote)
        elif isinstance(thing, path.BasePath):
            out = Writer(StringIO())
            thing = thing.realize(path_vars, shelly)
            escaped = out.write(thing, syntax, self.shell.inner_quote_info)

            thing = out.stream.getvalue()
            if shelly and escaped:
                thing = self.shell.wrap_quotes(thing)
            self.write_literal(thing)
        else:
            raise TypeError(type(thing))

        return escaped

    def write_each(self, things, syntax, delim=safe_str.literal(' '),
                   prefix=None, suffix=None):
        for i in iterutils.tween(things, delim, prefix, suffix):
            self.write(i, syntax)

    def write_shell(self, thing, syntax=Syntax.shell, can_wrap=False):
        if ( can_wrap and isinstance(thing, shell.shell_list) and
             platform_info().family == 'windows' ):
            prefix = safe_str.shell_literal('cmd /s /c "')
            suffix = safe_str.shell_literal('"')
        else:
            prefix = suffix = None
        self.write_each(iterutils.iterate(thing), syntax, prefix=prefix,
                        suffix=suffix)


class Variable(safe_str.safe_string_ops):
    def __init__(self, name):
        self.name = re.sub(r'\W', '_', name)

    def use(self):
        return safe_str.literal('${{{}}}'.format(self.name))

    def _safe_str(self):
        return self.use()

    def __str__(self):
        raise NotImplementedError()

    def __repr__(self):
        return repr(self.use())

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, rhs):
        return self.name == rhs.name

    def __ne__(self, rhs):
        return not (self == rhs)


def var(v):
    return v if isinstance(v, Variable) else Variable(v)


path_vars = {
    path.Root.srcdir  : Variable('srcdir'),
    path.Root.builddir: None,
}
path_vars.update({i: Variable(i.name) for i in path.InstallRoot})

# Only use destdir on platforms that actually support it (e.g. not Windows).
if platform_info().destdir:
    path_vars[path.DestDir.destdir] = Variable('DESTDIR')


class NinjaFile(object):
    Section = Section

    def __init__(self, bfgfile):
        self._bfgfile = bfgfile

        self._min_version = None
        self._var_table = set()
        self._variables = {i: [] for i in Section}

        self._rules = OrderedDict()

        self._builds = []
        self._build_outputs = set()
        self._defaults = []

    def min_version(self, version):
        version = Version(version)
        if self._min_version is None or version > self._min_version:
            self._min_version = version

    def variable(self, name, value, section=Section.other, exist_ok=False):
        name = var(name)
        if self.has_variable(name):
            if not exist_ok:
                raise ValueError('variable {!r} already exists'.format(name))
        else:
            self._var_table.add(name)
            value = self._convert_args(value)
            self._variables[section].append((name, value))
        return name

    def cmd_var(self, cmd):
        return self.variable(cmd.command_var, cmd.command, Section.command,
                             exist_ok=True)

    def has_variable(self, name):
        return var(name) in self._var_table

    def rule(self, name, command, depfile=None, deps=None, description=None,
             generator=False, pool=None, restat=False):
        command = self._convert_args(command)

        if pool is not None:
            if pool == 'console':
                self.min_version('1.5')
            else:
                raise ValueError('unknown pool {!r}'.format(pool))

        if re.search(r'\W', name):
            raise ValueError('rule name contains invalid characters')

        if self.has_rule(name):
            raise ValueError('rule {!r} already exists'.format(name))

        self._rules[name] = Rule(command, depfile, deps, description,
                                 generator, pool, restat)

    def has_rule(self, name):
        return name in self._rules

    @staticmethod
    def _output_str(name):
        out = Writer(StringIO())
        out.write(name, Syntax.output)
        return out.stream.getvalue()

    def build(self, output, rule, inputs=None, implicit=None, order_only=None,
              variables=None):
        if rule != 'phony' and not self.has_rule(rule):
            raise ValueError('unknown rule {!r}'.format(rule))

        variables = {var(k): self._convert_args(v) for k, v in
                     iteritems(variables or {})}

        outputs = iterutils.listify(output)
        for i in outputs:
            out = self._output_str(i)
            if self.has_build(out):
                raise ValueError('build for {!r} already exists'.format(out))
            self._build_outputs.add(out)
        self._builds.append(Build(
            outputs, rule, iterutils.listify(inputs),
            iterutils.listify(implicit), iterutils.listify(order_only),
            variables
        ))

    def has_build(self, name):
        return name in self._build_outputs

    def default(self, paths):
        self._defaults.extend(iterutils.iterate(paths))

    def _convert_args(self, args):
        if iterutils.isiterable(args):
            return Command.convert_args(args, self.cmd_var)
        return args

    def _write_variable(self, out, name, value, syntax=Syntax.shell, indent=0,
                        can_wrap=False):
        out.write_literal(('  ' * indent) + name.name + ' = ')
        out.write_shell(value, syntax, can_wrap)
        out.write_literal('\n')

    def _write_rule(self, out, name, rule):
        out.write_literal('rule ' + name + '\n')

        self._write_variable(out, var('command'), rule.command, indent=1,
                             can_wrap=True)
        if rule.depfile:
            self._write_variable(out, var('depfile'), rule.depfile, indent=1)
        if rule.deps:
            self._write_variable(out, var('deps'), rule.deps, indent=1)
        if rule.description:
            self._write_variable(out, var('description'), rule.description,
                                 indent=1, syntax=Syntax.clean)
        if rule.generator:
            self._write_variable(out, var('generator'), '1', indent=1)
        if rule.pool:
            self._write_variable(out, var('pool'), rule.pool, indent=1)
        if rule.restat:
            self._write_variable(out, var('restat'), '1', indent=1)

    def _write_build(self, out, build):
        out.write_literal('build ')
        out.write_each(build.outputs, Syntax.output)
        out.write_literal(': ' + build.rule)

        lit = safe_str.literal
        out.write_each(build.inputs, Syntax.input, prefix=lit(' '))
        out.write_each(build.implicit, Syntax.input, prefix=lit(' | '))
        out.write_each(build.order_only, Syntax.input, prefix=lit(' || '))
        out.write_literal('\n')

        if build.variables:
            desc_var = var('description')
            for k, v in iteritems(build.variables):
                syntax = Syntax.clean if k == desc_var else Syntax.shell
                self._write_variable(out, k, v, indent=1, syntax=syntax)

    def write(self, out):
        out = Writer(out)
        out.write_literal(_comment_tmpl.format(self._bfgfile) + '\n\n')

        if self._min_version:
            self._write_variable(
                out, var('ninja_required_version'), str(self._min_version)
            )
            out.write_literal('\n')

        for section in Section:
            # The built-in paths don't need shell quoting because they're used
            # by other paths, which *are* quoted.
            syntax = Syntax.clean if section == Section.path else Syntax.shell
            for name, value in self._variables[section]:
                self._write_variable(out, name, value, syntax)
            if self._variables[section]:
                out.write_literal('\n')

        for name, rule in iteritems(self._rules):
            self._write_rule(out, name, rule)
            out.write_literal('\n')

        for build in self._builds:
            self._write_build(out, build)
            out.write_literal('\n')

        if self._defaults:
            out.write_literal('default ')
            out.write_each(self._defaults, Syntax.input)
            out.write_literal('\n')



from six.moves import cStringIO as StringIO

from . import builtin
from .. import safe_str
from .. import shell
from ..backends.make import writer as make
from ..backends.ninja import writer as ninja
from ..build_inputs import build_input
from ..file_types import Node
from ..iterutils import first, iterate
from ..shell import posix as pshell


@build_input('tests')
class TestInputs(object):
    def __init__(self, build_inputs, env):
        self.tests = []
        self.extra_deps = []

    def __nonzero__(self):
        return self.__bool__()

    def __bool__(self):
        return bool(self.tests)


class Test(object):
    def __init__(self, build, env, cmd, environment, driver):
        # Ensure that bare Node objects are treated as a list of args instead
        # of a literal command line (the former has shell-characters escaped).
        if isinstance(cmd, Node):
            cmd = [cmd]
        wrap = not driver or driver.wrap_children
        self.cmd = env.run_arguments(cmd) if wrap else cmd

        self.inputs = [i for i in iterate(cmd)
                       if isinstance(i, Node) and i.creator]
        self.env = environment

        primary = first(cmd)
        if isinstance(primary, Node) and primary.creator:
            build['defaults'].remove(primary)
        (driver or build['tests']).tests.append(self)


class TestCase(Test):
    def __init__(self, build, env, cmd, environment={}, driver=None):
        if driver and environment:
            raise TypeError("only one of 'driver' and 'environment' may be " +
                            "specified")
        Test.__init__(self, build, env, cmd, environment, driver)


class TestDriver(Test):
    def __init__(self, build, env, cmd, environment={}, parent=None,
                 wrap_children=False):
        if parent and environment:
            raise TypeError("only one of 'parent' and 'environment' may be " +
                            "specified")

        Test.__init__(self, build, env, cmd, environment, parent)
        self.tests = []
        self.wrap_children = wrap_children


@builtin.function('build_inputs', 'env')
def test(build, env, cmd, **kwargs):
    return TestCase(build, env, cmd, **kwargs)


@builtin.function('build_inputs', 'env')
def test_driver(build, env, cmd, **kwargs):
    return TestDriver(build, env, cmd, **kwargs)


@builtin.function('build_inputs')
def test_deps(build, *args):
    if len(args) == 0:
        raise ValueError('expected at least one argument')
    build['tests'].extra_deps.extend(args)


def _build_commands(tests, writer, shell, local_env, collapse=False):
    def command(test, args=[]):
        subcmd = local_env(test.env, test.cmd) + args

        if collapse:
            out = writer(StringIO())
            out.write_shell(subcmd)
            s = out.stream.getvalue()
            if len(subcmd) > 1:
                s = shell.quote(s)
            return safe_str.literal(s)
        return subcmd

    cmd, deps = [], []
    for i in tests:
        deps.extend(i.inputs)
        if isinstance(i, TestDriver):
            args, more_deps = _build_commands(
                i.tests, writer, shell, local_env, True
            )
            cmd.append(command(i, args))
            deps.extend(more_deps)
        else:
            cmd.append(command(i))
    return cmd, deps


@make.post_rule
def make_test_rule(build_inputs, buildfile, env):
    tests = build_inputs['tests']
    if not tests:
        return

    recipe, deps = _build_commands(
        tests.tests, make.Writer, pshell, pshell.local_env
    )

    buildfile.rule(
        target='tests',
        deps=deps + tests.extra_deps,
        phony=True
    )
    buildfile.rule(
        target='test',
        deps='tests',
        recipe=recipe,
        phony=True
    )


@ninja.post_rule
def ninja_test_rule(build_inputs, buildfile, env):
    tests = build_inputs['tests']
    if not tests:
        return

    try:
        local_env = shell.local_env
    except AttributeError:
        local_env = env.tool('setenv')

    commands, deps = _build_commands(
        tests.tests, ninja.Writer, shell, local_env
    )

    buildfile.build(
        output='tests',
        rule='phony',
        inputs=deps + tests.extra_deps
    )
    ninja.command_build(
        buildfile, env,
        output='test',
        inputs='tests',
        command=shell.join_lines(commands),
        console=True, phony=True
    )



import os.path
import re

from . import builtin
from .find import find
from .. import options as opts
from ..exceptions import PackageResolutionError, PackageVersionError
from ..file_types import Directory, Executable
from ..iterutils import default_sentinel, iterate, uniques
from ..languages import known_langs
from ..objutils import objectify
from ..packages import CommonPackage, Framework, Package, PackageKind
from ..path import Path, Root
from ..shell import which
from ..versioning import check_version, SpecifierSet, Version


class BoostPackage(CommonPackage):
    def __init__(self, name, format, version, *args, **kwargs):
        name = 'boost({})'.format(','.join(iterate(name)))
        CommonPackage.__init__(self, name, format, *args, **kwargs)
        self.version = version


@builtin.function('build_inputs', 'env')
@builtin.type(Package)
def package(build, env, name, version=None, lang=default_sentinel,
            kind=PackageKind.any.name, headers=None, libs=default_sentinel):
    version = objectify(version or '', SpecifierSet)
    kind = PackageKind[kind]

    if lang is default_sentinel:
        guessed_langs = uniques(
            known_langs.fromext(os.path.splitext(i)[1], 'header')
            for i in iterate(headers)
        )
        if len(guessed_langs) == 1 and guessed_langs[0] is not None:
            lang = guessed_langs[0]
        else:
            lang = build['project']['lang']

    return env.builder(lang).packages.resolve(name, version, kind, headers,
                                              libs)


@builtin.function('env')
@builtin.type(Executable)
def system_executable(env, name, format=None):
    return Executable(
        Path(which([[name]], env.variables, resolve=True)[0], Root.absolute),
        format or env.target_platform.object_format
    )


@builtin.function('env')
def framework(env, name, suffix=None):
    if not env.target_platform.has_frameworks:
        raise PackageResolutionError("{} platform doesn't support frameworks"
                                     .format(env.target_platform.name))

    framework = Framework(name, suffix)
    return CommonPackage(framework.full_name,
                         env.target_platform.object_format,
                         link_options=opts.option_list(opts.lib(framework)))


def _boost_version(header, required_version):
    version_hpp = header.path.append('boost').append('version.hpp')
    with open(version_hpp.string()) as f:
        for line in f:
            m = re.match(r'#\s*define\s+BOOST_LIB_VERSION\s+"([\d_]+)"', line)
            if m:
                version = Version(m.group(1).replace('_', '.'))
                check_version(version, required_version, 'boost',
                              PackageVersionError)
                return version
    raise PackageVersionError('unable to parse "boost/version.hpp"')


@builtin.function('env')
def boost_package(env, name=None, version=None):
    def getdir(name, root, default):
        d = env.getvar(name, os.path.join(root, default) if root else None)
        return Path(d, Root.absolute) if d else None

    version = objectify(version or '', SpecifierSet)
    pkg = env.builder('c++').packages
    version_hpp = 'boost/version.hpp'

    root = env.getvar('BOOST_ROOT')
    incdir = getdir('BOOST_INCLUDEDIR', root, 'include')
    libdir = getdir('BOOST_LIBRARYDIR', root, 'lib')
    header = None

    if env.target_platform.family == 'windows':
        if not env.builder('c++').auto_link:  # pragma: no cover
            # XXX: Don't require auto-link.
            raise PackageResolutionError('Boost on Windows requires auto-link')

        # On Windows, check the default install location, which is structured
        # differently from other install locations.
        if not incdir and env.host_platform.family == 'windows':
            dirs = find(env, r'C:\Boost\include', 'boost-*', type='d',
                        flat=True)
            if dirs:
                try:
                    header = pkg.header(version_hpp, [max(dirs)])
                    libdir = Path(r'C:\Boost\lib', Root.absolute)
                except PackageResolutionError:
                    pass

        if header is None:
            header = pkg.header(version_hpp, [incdir] if incdir else None)

        compile_options = opts.option_list(opts.include_dir(header))
        link_options = opts.option_list()
        if libdir:
            link_options.append(opts.lib_dir( Directory(libdir) ))
    else:
        header = pkg.header(version_hpp, [incdir] if incdir else None)

        dirs = [libdir] if libdir else None
        libs = (pkg.library('boost_' + i, search_dirs=dirs)
                for i in iterate(name))

        compile_options = opts.option_list()
        link_options = opts.option_list()
        if env.target_platform.family == 'posix' and 'thread' in iterate(name):
            compile_options.append(opts.pthread())
            link_options.append(opts.pthread())

        compile_options.append(opts.include_dir(header))
        link_options.extend(opts.lib(i) for i in libs)

    return BoostPackage(
        name, env.builder('c++').object_format,
        _boost_version(header, version),
        compile_options, link_options
    )



from ..backends.make import writer as make


@make.post_rule
def make_clean_rule(build_inputs, buildfile, env):
    rm = env.tool('rm')
    buildfile.rule(target='clean', recipe=[
        rm(i.path for i in build_inputs.targets())
    ], phony=True)



import re

from .. import shell
from ..versioning import detect_version
from ..path import abspath


class LdLinker(object):
    def __init__(self, builder, env, command, version_output):
        self.builder = builder
        self.env = env
        self.command = command

        if 'GNU ld' in version_output:
            self.brand = 'bfd'
            self.version = detect_version(version_output)
        elif 'GNU gold' in version_output:
            self.brand = 'gold'
            self.version = detect_version(version_output, post='$',
                                          flags=re.MULTILINE)
        else:
            self.brand = 'unknown'
            self.version = None

    @property
    def lang(self):
        return self.builder.lang

    @property
    def family(self):
        return self.builder.family

    @property
    def flavor(self):
        return 'ld'

    def search_dirs(self, sysroot='/', strict=False):
        try:
            output = self.env.execute(
                self.command + ['--verbose'], stdout=shell.Mode.pipe,
                stderr=shell.Mode.devnull
            )
            search_dirs = [i.group(1) for i in re.finditer(
                r'SEARCH_DIR\("((?:[^"\\]|\\.)*)"\)', output)
            ]
            return [abspath(sysroot.rstrip('/') + i[1:] if i[0] == '=' else i)
                    for i in search_dirs]
        except (OSError, shell.CalledProcessError):
            if strict:
                raise
            return []



from .basepath import BasePath, InstallRoot
from .core import Platform
from .host import HostPlatform
from .target import TargetPlatform


class WindowsPath(BasePath):
    def _localize_path(self, path):
        return path.replace('/', '\\')


class WindowsPlatform(Platform):
    @property
    def _triplet_sys_abi(self):
        return 'win32'

    @property
    def family(self):
        return 'windows'

    Path = WindowsPath


class WindowsHostPlatform(HostPlatform, WindowsPlatform):
    @property
    def include_dirs(self):
        # Windows doesn't have standard include dirs; for MSVC, this will get
        # pulled in from the $INCLUDE environment variable.
        return []

    @property
    def lib_dirs(self):
        # Windows doesn't have standard library dirs; for MSVC, this will get
        # pulled in from the $LIB environment variable.
        return []

    @property
    def destdir(self):
        return False


class WindowsTargetPlatform(TargetPlatform, WindowsPlatform):
    _package_map = {
        'gl': 'opengl32',
        'glu': 'glu32',
        'glut': 'glut32',
    }

    @property
    def object_format(self):
        return 'coff'

    @property
    def executable_ext(self):
        return '.exe'

    @property
    def shared_library_ext(self):
        return '.dll'

    @property
    def has_import_library(self):
        return True

    @property
    def has_versioned_library(self):
        return False

    @property
    def has_frameworks(self):
        return False

    @property
    def install_dirs(self):
        return {
            InstallRoot.prefix     : None,
            InstallRoot.exec_prefix: WindowsPath('', InstallRoot.prefix),
            InstallRoot.bindir     : WindowsPath('', InstallRoot.exec_prefix),
            InstallRoot.libdir     : WindowsPath('', InstallRoot.exec_prefix),
            InstallRoot.includedir : WindowsPath('', InstallRoot.prefix),
        }



import os.path
import re

from . import *


class TestFilesWithSpaces(IntegrationTest):
    def __init__(self, *args, **kwargs):
        IntegrationTest.__init__(self, 'files_with_spaces', *args, **kwargs)

    def test_build(self):
        self.build(executable('quad damage'))
        self.assertOutput([executable('quad damage')], 'QUAD DAMAGE!\n')

    def test_build_sub_dir(self):
        self.build(executable('another file'))
        self.assertOutput([executable('another file')], 'hello from sub dir\n')

    @only_if_backend('make', hide=True)
    def test_dir_sentinels(self):
        self.build(executable('another file'))
        self.assertTrue(os.path.isfile('another file.int/sub dir/.dir'))

    @skip_if_backend('msbuild')
    def test_script(self):
        assertRegex(self, self.build('script'),
                    re.compile('^hello, world!$', re.MULTILINE))



from . import builtin
from ..arguments.parser import add_user_argument


@builtin.getter('argv')
def argv(_argv):
    return _argv


@builtin.function('parser', context='options')
def argument(parser, *args, **kwargs):
    names = ['--' + i for i in args]
    add_user_argument(parser, *names, **kwargs)
